import os
import sys
import petsc4py
import numpy as np
import pandas as pd
import ruamel.yaml as yaml

from operator import itemgetter
from scipy.interpolate import interp1d

petsc4py.init(sys.argv)
MPIrank = petsc4py.PETSc.COMM_WORLD.Get_rank()


class ReadYaml(object):
    """
    Class for reading simulation input file and initialising model parameters.

    Definition of input parameters is provided in the `User Documentation <https://gospl.readthedocs.io/en/latest/inputfile.html>`_
    """

    def __init__(self, filename):
        """
        Parsing YAML file.

        :arg filename: input filename (.yml YAML file)
        """

        if self.showlog:
            self.log = petsc4py.PETSc.Log()
            self.log.begin()

        # Check input file exists
        try:
            with open(filename) as finput:
                pass
        except IOError:
            print("Unable to open file: ", filename, flush=True)
            raise IOError("The input file is not found...")

        # Open YAML file
        with open(filename, "r") as finput:
            self.input = yaml.load(finput, Loader=yaml.Loader)

        if MPIrank == 0 and "name" in self.input.keys() and self.verbose:
            print(
                "The following model will be run:     {}".format(self.input["name"]),
                flush=True,
            )

        # Read simulation parameters
        self._readDomain()
        self._readTime()
        self._readSPL()
        self._readHillslope()
        self._readSealevel()
        self._readTectonic()
        self._readRain()
        self._readCompaction()

        if self.raindata is not None:
            if self.rStep > 0:
                self.raindata = self.raindata[
                    self.raindata["start"] >= self.tStart + self.rStep * self.tout
                ]
                self.raindata.reset_index(drop=True, inplace=True)
                self.rainNb = len(self.raindata)

        self._readBackwardPaleo()
        self._readOut()
        self._readForcePaleo()

        self.radius = 6378137.0
        self.gravity = 9.81
        self.tNow = self.tStart
        self.saveTime = self.tNow
        if self.strat > 0:
            self.saveStrat = self.tNow + self.strat
        else:
            self.saveStrat = self.tEnd + self.tout

        return

    def _readDomain(self):
        """
        Read domain definition, boundary conditions and flow direction parameters.
        """

        try:
            domainDict = self.input["domain"]
        except KeyError:
            print(
                "Key 'domain' is required and is missing in the input file!", flush=True
            )
            raise KeyError("Key domain is required in the input file!")

        # try:
        #     self.radius = domainDict["radius"]
        # except KeyError:
        #     self.radius = 6378137.0

        try:
            self.flowDir = domainDict["flowdir"]
        except KeyError:
            self.flowDir = 6

        try:
            self.flowExp = domainDict["flowexp"]
        except KeyError:
            self.flowExp = 0.42

        try:
            meshFile = domainDict["npdata"]
        except KeyError:
            print(
                "Key 'npdata' is required and is missing in the 'domain' declaration!",
                flush=True,
            )
            raise KeyError("Compressed numpy dataset definition is not defined!")

        self.meshFile = meshFile + ".npz"

        try:
            with open(self.meshFile) as meshfile:
                meshfile.close()

        except IOError:
            print("Unable to open numpy dataset: {}".format(self.meshFile), flush=True)
            raise IOError("The numpy dataset is not found...")

        try:
            self.fast = domainDict["fast"]
        except KeyError:
            self.fast = False

        try:
            self.backward = domainDict["backward"]
        except KeyError:
            self.backward = False

        try:
            self.interp = domainDict["interp"]
        except KeyError:
            self.interp = 1

        self._extraDomain()

        return

    def _extraDomain(self):
        """
        Read domain additional information.
        """

        domainDict = self.input["domain"]

        try:
            self.overlap = domainDict["overlap"]
        except KeyError:
            self.overlap = 1

        try:
            dataFile = domainDict["nperodep"]
            self.dataFile = dataFile + ".npz"
            with open(self.dataFile) as dataFile:
                dataFile.close()
        except KeyError:
            self.dataFile = None

        try:
            strataFile = domainDict["npstrata"]
            self.strataFile = strataFile + ".npz"
            with open(self.strataFile) as strataFile:
                strataFile.close()
        except KeyError:
            self.strataFile = None

        return

    def _readTime(self):
        """
        Read simulation time declaration.
        """

        try:
            timeDict = self.input["time"]
        except KeyError:
            print(
                "Key 'time' is required and is missing in the input file!", flush=True
            )
            raise KeyError("Key time is required in the input file!")

        try:
            self.tStart = timeDict["start"]
        except KeyError:
            print(
                "Key 'start' is required and is missing in the 'time' declaration!",
                flush=True,
            )
            raise KeyError("Simulation start time needs to be declared.")

        try:
            self.tEnd = timeDict["end"]
        except KeyError:
            print(
                "Key 'end' is required and is missing in the 'time' declaration!",
                flush=True,
            )
            raise KeyError("Simulation end time needs to be declared.")

        if self.tEnd <= self.tStart:
            raise ValueError("Simulation end/start times do not make any sense!")

        try:
            self.dt = timeDict["dt"]
        except KeyError:
            print(
                "Key 'dt' is required and is missing in the 'time' declaration!",
                flush=True,
            )
            raise KeyError("Simulation discretisation time step needs to be declared.")

        try:
            self.tout = timeDict["tout"]
        except KeyError:
            self.tout = self.tEnd - self.tStart
            print(
                "Output time interval 'tout' has been set to {} years".format(
                    self.tout
                ),
                flush=True,
            )

        self._addTime(timeDict)

        return

    def _addTime(self, timeDict):
        """
        Read additional time parameters.
        """

        try:
            self.rStep = timeDict["rstep"]
        except KeyError:
            self.rStep = 0

        if self.tout < self.dt:
            self.tout = self.dt
            print(
                "Output time interval was changed to {} years to match the time step dt".format(
                    self.dt
                ),
                flush=True,
            )

        try:
            self.tecStep = timeDict["tec"]
        except KeyError:
            self.tecStep = self.tout

        try:
            self.strat = timeDict["strat"]
        except KeyError:
            self.strat = 0

        if self.tout < self.tecStep:
            self.tecStep = self.tout
            print(
                "Output time interval and tectonic forcing time step \
                 have been adjusted to match each others.",
                flush=True,
            )

        if self.tout < self.strat:
            self.strat = self.tout
            print(
                "Output time interval and stratal time step \
                 have been adjusted to match each others.",
                flush=True,
            )

        if self.tecStep > 0:
            if self.tout % self.tecStep != 0:
                print(
                    "When declaring tectonic time interval, the value should be divisible by the output time interval.",
                    flush=True,
                )
                raise ValueError("Tectonic time interval definition is wrong!")

        if self.strat > 0:
            if self.tout % self.strat != 0:
                print(
                    "When declaring stratal time interval, the value should be divisible by the output time interval.",
                    flush=True,
                )
                raise ValueError("Stratal time interval definition is wrong!")
            self.stratNb = int((self.tEnd - self.tStart) / self.strat) + 1
        else:
            self.stratNb = 0

        return

    def _readSPL(self):
        """
        Read surface processes bedrock parameters.
        """

        try:
            splDict = self.input["spl"]
            try:
                self.K = splDict["K"]
            except KeyError:
                print(
                    "When using the Surface Process Model definition of coefficient Kb is required.",
                    flush=True,
                )
                raise ValueError("Surface Process Model: Kb coefficient not found.")
            try:
                self.waterfill = splDict["wfill"]
            except KeyError:
                self.waterfill = 100.0
            try:
                self.sedfill = splDict["sfill"]
            except KeyError:
                self.sedfill = 50.0
            try:
                self.coeffd = splDict["d"]
            except KeyError:
                self.coeffd = 0.0
            # try:
            #     # `wght` is the percentage of upstream sediment flux
            #     # that will be deposited on each cell...
            #     self.wght = splDict["wght"]
            #     if self.wght >= 1.0:
            #         self.wght = 0.999
            # except KeyError:
            #     self.wght = 0.0

        except KeyError:
            self.K = 1.0e-12
            self.waterfill = 100.0
            self.sedfill = 50.0
            # self.wght = 0.0
            self.coeffd = 0.0

        return

    def _readHillslope(self):
        """
        Read hillslope parameters.
        """

        try:
            hillDict = self.input["diffusion"]

            try:
                self.Cda = hillDict["hillslopeKa"]
            except KeyError:
                print(
                    "When declaring diffusion processes, the coefficient hillslopeKa is required.",
                    flush=True,
                )
                raise ValueError(
                    "Hillslope: Cd coefficient for aerial environment not found."
                )
            try:
                self.Cdm = hillDict["hillslopeKm"]
            except KeyError:
                print(
                    "When declaring diffusion processes, the coefficient hillslopeKm is required.",
                    flush=True,
                )
                raise ValueError(
                    "Hillslope: Cd coefficient for marine environment not found."
                )
            try:
                self.sedimentK = hillDict["sedK"]
            except KeyError:
                self.sedimentK = 10000.0
            try:
                self.sedimentKf = hillDict["sedKf"]
            except KeyError:
                self.sedimentKf = 20000.0
            try:
                self.sedimentKw = hillDict["sedKw"]
            except KeyError:
                self.sedimentKw = 30000.0
            try:
                self.diffstep = hillDict["dstep"]
            except KeyError:
                self.diffstep = 3
        except KeyError:
            self.Cda = 0.0
            self.Cdm = 0.0
            self.sedimentK = 10000.0
            self.sedimentKf = 20000.0
            self.sedimentKw = 30000.0
            self.diffstep = 3

        return

    def _readSealevel(self):
        """
        Define sealevel evolution.
        """

        seafile = None
        sealevel = 0.0
        self.seafunction = None
        try:
            seaDict = self.input["sea"]
            try:
                sealevel = seaDict["position"]
                try:
                    seafile = seaDict["curve"]
                except KeyError:
                    seafile = None
            except KeyError:
                try:
                    seafile = seaDict["curve"]
                except KeyError:
                    seafile = None
        except KeyError:
            sealevel = 0.0

        if seafile is not None:
            try:
                with open(seafile) as fsea:
                    fsea.close()
                    try:
                        seadata = pd.read_csv(
                            seafile,
                            sep=r",",
                            engine="c",
                            header=None,
                            na_filter=False,
                            dtype=np.float64,
                            low_memory=False,
                        )

                    except ValueError:
                        try:
                            seadata = pd.read_csv(
                                seafile,
                                sep=r"\s+",
                                engine="c",
                                header=None,
                                na_filter=False,
                                dtype=np.float64,
                                low_memory=False,
                            )

                        except ValueError:
                            print(
                                "The sea-level file is not well formed: it should be comma or tab separated",
                                flush=True,
                            )
                            raise ValueError("Wrong formating of sea-level file.")
            except IOError:
                print("Unable to open file: ", seafile)
                raise IOError("The sealevel file is not found...")

            seadata[1] += sealevel
            if seadata[0].min() > self.tStart:
                tmpS = []
                tmpS.insert(0, {0: self.tStart, 1: seadata[1].iloc[0]})
                seadata = pd.concat([pd.DataFrame(tmpS), seadata], ignore_index=True)
            if seadata[0].max() < self.tEnd:
                tmpE = []
                tmpE.insert(0, {0: self.tEnd, 1: seadata[1].iloc[-1]})
                seadata = pd.concat([seadata, pd.DataFrame(tmpE)], ignore_index=True)
            self.seafunction = interp1d(seadata[0], seadata[1], kind="linear")
        else:
            year = np.linspace(self.tStart, self.tEnd + self.dt, num=11, endpoint=True)
            seaval = np.full(len(year), sealevel)
            self.seafunction = interp1d(year, seaval, kind="linear")

        return

    def _storeTectonic(self, k, tecStart, zMap, tMap, tStep, tEnd, tecdata):
        """
        Record tectonic conditions.

        :arg k: tectonic event number
        :arg tecStart: tectonic event start time
        :arg zMap: horizontal tectonic displacement file
        :arg tMap: vertical tectonic displacement file
        :arg tStep: tectonic time step
        :arg tEnd: tectonic event end time
        :arg tecdata: pandas dataframe storing each tectonic event

        :return: appended tecdata
        """

        if tMap is not None:
            if self.meshFile != tMap:
                try:
                    with open(tMap) as tecfile:
                        tecfile.close()

                except IOError:
                    print("Unable to open tectonic file: {}".format(tMap), flush=True)
                    raise IOError(
                        "The tectonic file {} is not found for climatic event {}.".format(
                            tMap, k
                        )
                    )
        else:
            tMap = "empty"

        if zMap is not None:
            if self.meshFile != zMap:
                try:
                    with open(zMap) as tecfile:
                        tecfile.close()

                except IOError:
                    print("Unable to open tectonic file: {}".format(zMap), flush=True)
                    raise IOError(
                        "The tectonic file {} is not found for climatic event {}.".format(
                            zMap, k
                        )
                    )
        else:
            zMap = "empty"

        if tMap == "empty" and zMap == "empty":
            print(
                "For each tectonic event a tectonic grid (mapH or mapV) is required.",
                flush=True,
            )
            raise ValueError("Tectonic event {} has no tectonic map (map).".format(k))

        tmpTec = []
        tmpTec.insert(0, {"start": tecStart, "tMap": tMap, "zMap": zMap})

        if k == 0:
            tecdata = pd.DataFrame(tmpTec, columns=["start", "tMap", "zMap"])
        else:
            tecdata = pd.concat(
                [tecdata, pd.DataFrame(tmpTec, columns=["start", "tMap", "zMap"])],
                ignore_index=True,
            )

        if tStep is not None:
            if tEnd is not None:
                tectime = tecStart + tStep
                while tectime < tEnd:
                    tmpTec = []
                    tmpTec.insert(0, {"start": tectime, "tMap": tMap, "zMap": zMap})
                    tecdata = pd.concat(
                        [
                            tecdata,
                            pd.DataFrame(tmpTec, columns=["start", "tMap", "zMap"]),
                        ],
                        ignore_index=True,
                    )
                    tectime = tectime + tStep

        return tecdata

    def _defineTectonic(self, k, tecSort, tecdata):
        """
        Define tectonic conditions.

        :arg k: tectonic event number
        :arg tecSort: sorted tectonic event
        :arg tecdata: pandas dataframe storing each tectonic event
        :return: appended tecdata
        """

        tecStart = None
        tEnd = None
        tStep = None
        tMap = None
        zMap = None

        try:
            tecStart = tecSort[k]["start"]
        except Exception:
            print("For each tectonic event a start time is required.", flush=True)
            raise ValueError("Tectonic event {} has no parameter start".format(k))
        try:
            tMap = tecSort[k]["mapH"] + ".npz"
        except Exception:
            pass
        try:
            zMap = tecSort[k]["mapV"] + ".npz"
        except Exception:
            pass
        try:
            tStep = self.tecStep
        except Exception:
            pass
        try:
            tEnd = tecSort[k]["end"]
        except Exception:
            pass

        tecdata = self._storeTectonic(k, tecStart, zMap, tMap, tStep, tEnd, tecdata)

        return tecdata

    def _readTectonic(self):
        """
        Parse tectonic forcing conditions.
        """

        tecdata = None
        try:
            tecDict = self.input["tectonic"]
            tecSort = sorted(tecDict, key=itemgetter("start"))
            for k in range(len(tecSort)):
                tecdata = self._defineTectonic(k, tecSort, tecdata)

            if tecdata["start"][0] > self.tStart:
                tmpTec = []
                tmpTec.insert(
                    0, {"start": self.tStart, "tMap": "empty", "zMap": "empty"}
                )
                tecdata = pd.concat(
                    [pd.DataFrame(tmpTec, columns=["start", "tMap", "zMap"]), tecdata],
                    ignore_index=True,
                )
            self.tecdata = tecdata[tecdata["start"] >= self.tStart]
            if self.rStep > 0:
                self.tecdata = tecdata[
                    tecdata["start"] >= self.tStart + self.rStep * self.tout
                ]
            self.tecdata.reset_index(drop=True, inplace=True)

        except KeyError:
            self.tecdata = None

        return

    def _defineRain(self, k, rStart, rMap, rUniform, raindata):
        """
        Define precipitation conditions.

        :arg k: precipitation event number
        :arg rStart: precipitation event start time
        :arg rMap: precipitation map file event
        :arg rUniform: precipitation uniform value event
        :arg raindata: pandas dataframe storing each precipitation event
        :return: appended raindata
        """

        if rMap is None and rUniform is None:
            print(
                "For each climate event a rainfall value (uniform) or a rainfall \
                grid (map) is required.",
                flush=True,
            )
            raise ValueError(
                "Climate event {} has no rainfall value (uniform) or a rainfall \
                map (map).".format(
                    k
                )
            )

        tmpRain = []
        if rMap is None:
            tmpRain.insert(
                0,
                {"start": rStart, "rUni": rUniform, "rMap": None, "rKey": None},
            )
        else:
            tmpRain.insert(
                0,
                {
                    "start": rStart,
                    "rUni": None,
                    "rMap": rMap[0] + ".npz",
                    "rKey": rMap[1],
                },
            )

        if k == 0:
            raindata = pd.DataFrame(tmpRain, columns=["start", "rUni", "rMap", "rKey"])
        else:
            raindata = pd.concat(
                [
                    raindata,
                    pd.DataFrame(tmpRain, columns=["start", "rUni", "rMap", "rKey"]),
                ],
                ignore_index=True,
            )

        return raindata

    def _readRain(self):
        """
        Parse rain forcing conditions.
        """

        raindata = None
        try:
            rainDict = self.input["climate"]
            rainSort = sorted(rainDict, key=itemgetter("start"))
            for k in range(len(rainSort)):
                rStart = None
                rUniform = None
                rMap = None
                try:
                    rStart = rainSort[k]["start"]
                except Exception:
                    print(
                        "For each climate event a start time is required.", flush=True
                    )
                    raise ValueError(
                        "Climate event {} has no parameter start".format(k)
                    )
                try:
                    rUniform = rainSort[k]["uniform"]
                except Exception:
                    pass
                try:
                    rMap = rainSort[k]["map"]
                except Exception:
                    pass

                if rMap is not None:
                    if self.meshFile != rMap[0] + ".npz":
                        try:
                            with open(rMap[0] + ".npz") as rainfile:
                                rainfile.close()

                        except IOError:
                            print(
                                "Unable to open rain file: {}.npz".format(rMap[0]),
                                flush=True,
                            )
                            raise IOError(
                                "The rain file {} is not found for climatic event {}.".format(
                                    rMap[0] + ".npz", k
                                )
                            )
                        mdata = np.load(rMap[0] + ".npz")
                        rainSet = mdata.files
                    else:
                        mdata = np.load(self.meshFile)
                        rainSet = mdata.files
                    try:
                        rainKey = mdata[rMap[1]]
                        if rainKey is not None:
                            pass
                    except KeyError:
                        print(
                            "Field name {} is missing from rain file {}.npz".format(
                                rMap[1], rMap[0]
                            ),
                            flush=True,
                        )
                        print(
                            "The following fields are available: {}".format(rainSet),
                            flush=True,
                        )
                        print("Check your rain file fields definition...", flush=True)
                        raise KeyError(
                            "Field name for rainfall is not defined correctly or does not exist!"
                        )

                raindata = self._defineRain(k, rStart, rMap, rUniform, raindata)

            if raindata["start"][0] > self.tStart:
                tmpRain = []
                tmpRain.insert(
                    0, {"start": self.tStart, "rUni": 0.0, "rMap": None, "rKey": None}
                )
                raindata = pd.concat(
                    [
                        pd.DataFrame(
                            tmpRain, columns=["start", "rUni", "rMap", "rKey"]
                        ),
                        raindata,
                    ],
                    ignore_index=True,
                )
            self.raindata = raindata[raindata["start"] >= self.tStart]
            self.raindata.reset_index(drop=True, inplace=True)
            self.rainNb = len(self.raindata)

        except KeyError:
            self.raindata = None

        return

    def _readBackwardPaleo(self):
        """
        Force model with backward paleomaps.
        """
        try:
            paleoDict = self.input["paleomap"]
            paleoSort = sorted(paleoDict, key=itemgetter("time"))
            for k in range(len(paleoSort)):
                pTime = None
                pMap = None
                try:
                    pTime = paleoSort[k]["time"]
                except Exception:
                    print("For each paleomap a given time is required.", flush=True)
                    raise ValueError("Paleomap {} has no parameter time".format(k))
                try:
                    pMap = paleoSort[k]["npdata"]
                except Exception:
                    pass

                if pMap is not None:
                    try:
                        with open(pMap + ".npz") as meshfile:
                            meshfile.close()

                    except IOError:
                        print(
                            "Unable to open numpy dataset: {}.npz".format(pMap),
                            flush=True,
                        )
                        raise IOError("The numpy dataset is not found...")

                tmpPaleo = []
                tmpPaleo.insert(0, {"time": pTime, "pMap": pMap + ".npz"})
                if k == 0:
                    paleodata = pd.DataFrame(tmpPaleo, columns=["time", "pMap"])
                else:
                    paleodata = pd.concat(
                        [paleodata, pd.DataFrame(tmpPaleo, columns=["time", "pMap"])],
                        ignore_index=True,
                    )

            self.paleodata = paleodata[paleodata["time"] >= self.tStart]
            if self.rStep > 0:
                self.paleodata = paleodata[
                    paleodata["start"] >= self.tStart + self.rStep * self.tout
                ]

            self.paleodata.reset_index(drop=True, inplace=True)
            self.paleoNb = len(self.paleodata)

        except KeyError:
            self.paleodata = None
            self.paleoNb = 0

        return

    def _readForcePaleo(self):
        """
        Get series of paleomaps to force the model through time.
        """

        try:
            fpaleoDict = self.input["forcepaleo"]

            try:
                self.forceDir = fpaleoDict["dir"]
                if not os.path.exists(self.forceDir):
                    print("Forcing paleo directory does not exist!", flush=True)
                    raise ValueError("Forcing paleo directory does not exist!")
                try:
                    forceNb = fpaleoDict["steps"]
                except KeyError:
                    print(
                        "New Paleomap loading frequency 'steps' is required",
                        flush=True,
                    )
                outNb = np.sum(forceNb)
                self.stepb = np.flip(np.arange(0, outNb, dtype=int))
                self.alpha = np.zeros(len(self.stepb))
                p = 0
                for k in range(len(forceNb)):
                    out_nb = forceNb[k] + 1
                    stepf = np.arange(1, out_nb, dtype=int)
                    self.alpha[p : p + len(stepf)] = stepf.astype(float) / (out_nb - 1)
                    p += len(stepf)
                self.forceStep = 0

                if self.rStep > 0:
                    steptime = np.arange(0, outNb, dtype=int) * self.tout + self.tStart
                    id = np.where(steptime == self.tStart + self.rStep * self.tout)[0]
                    if len(id) == 0:
                        raise steptime(
                            "Something went wrong with the restart time, it needs to be related to the forcing step."
                        )
                    self.forceStep = id[0]

            except Exception:
                print(
                    "A directory is required to force the model with paleodata.",
                    flush=True,
                )
                raise ValueError("forcepaleo key requires a directory")

        except KeyError:
            self.forceDir = None
            self.forceStep = -1

        return

    def _readCompaction(self):
        """
        Read compaction parameters.
        """

        try:
            compDict = self.input["compaction"]
            try:
                self.phi0s = compDict["phis"]
            except KeyError:
                self.phi0s = 0.49

            try:
                self.phi0f = compDict["phif"]
            except KeyError:
                self.phi0f = 0.63

            try:
                self.phi0w = compDict["phiw"]
            except KeyError:
                self.phi0w = 0.63

            try:
                self.phi0c = compDict["phic"]
            except KeyError:
                self.phi0c = 0.65

        except KeyError:
            self.phi0s = 0.49
            self.phi0f = 0.63
            self.phi0w = 0.63
            self.phi0c = 0.65

        self._extraCompaction()

        return

    def _extraCompaction(self):
        """
        Read compaction additional parameters.
        """

        try:
            compDict = self.input["compaction"]
            try:
                self.z0s = compDict["z0s"]
            except KeyError:
                self.z0s = 3700.0

            try:
                self.z0f = compDict["z0f"]
            except KeyError:
                self.z0f = 1960.0

            try:
                self.z0w = compDict["z0w"]
            except KeyError:
                self.z0w = 1960.0

            try:
                self.z0c = compDict["z0c"]
            except KeyError:
                self.z0c = 2570.0

        except KeyError:
            self.z0s = 3700.0
            self.z0f = 1960.0
            self.z0w = 1960.0
            self.z0c = 2570.0

        return

    def _readOut(self):
        """
        Parse output directory.
        """

        try:
            outDict = self.input["output"]
            try:
                self.outputDir = outDict["dir"]
            except KeyError:
                self.outputDir = "output"
            try:
                self.makedir = outDict["makedir"]
            except KeyError:
                self.makedir = True
        except KeyError:
            self.outputDir = "output"
            self.makedir = True

        if self.rStep > 0:
            self.makedir = False

        return



import os
import gc
import sys
import petsc4py
import numpy as np
import numpy_indexed as npi

from mpi4py import MPI
from time import process_time

if "READTHEDOCS" not in os.environ:
    from gospl._fortran import fillpit
    from gospl._fortran import setmaxnb
    from gospl._fortran import mfdreceivers
    from gospl._fortran import seaparams
    from gospl._fortran import distsea
    from gospl._fortran import distexcess
    from gospl._fortran import distocean
    from gospl._fortran import mfdrcvs
    from gospl._fortran import sethillslopecoeff

petsc4py.init(sys.argv)
MPIrank = petsc4py.PETSc.COMM_WORLD.Get_rank()
MPIcomm = petsc4py.PETSc.COMM_WORLD


class SEDMesh(object):
    """
    This class encapsulates all the functions related to sediment transport, production and deposition. `gospl` has the ability to track three types of clastic sediment size and one type of carbonate (still under development). The following processes are considered:

    - inland river deposition in depressions and enclosed sea
    - marine deposition at river mouth
    - hillslope processes in both marine and inland areas

    .. note::
        Most of these functions are ran in parallel using the underlying PETSc library.

    """

    def __init__(self, *args, **kwargs):
        """
        The initialisation of `SEDMesh` class consists in the declaration of several PETSc vectors.
        """

        # Petsc vectors
        self.tmp = self.hGlobal.duplicate()
        self.tmpL = self.hLocal.duplicate()
        self.tmp1 = self.hGlobal.duplicate()
        self.Qs = self.hGlobal.duplicate()
        self.QsL = self.hLocal.duplicate()

        self.vSed = self.hGlobal.duplicate()
        self.vSedLocal = self.hLocal.duplicate()
        if self.stratNb > 0:
            if self.stratF is not None:
                self.vSedf = self.hGlobal.duplicate()
                self.vSedfLocal = self.hLocal.duplicate()
            if self.stratW is not None:
                self.vSedw = self.hGlobal.duplicate()
                self.vSedwLocal = self.hLocal.duplicate()
            if self.carbOn:
                self.vSedc = self.hGlobal.duplicate()
                self.vSedcLocal = self.hLocal.duplicate()

        # Get the maximum number of neighbours on the mesh
        maxnb = np.zeros(1, dtype=np.int64)
        maxnb[0] = setmaxnb(self.lpoints)
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, maxnb, op=MPI.MAX)
        self.maxnb = maxnb[0]

        return

    def getSedFlux(self):
        """
        This function computes sediment flux in cubic metres per year from incoming rivers. Like for the computation of the flow discharge and erosion rates, the sediment flux is solved by an implicit time integration method, the matrix system is the one obtained from the receiver distributions over the unfilled elevation mesh for the flow discharge (`fMat`).

        The PETSc *scalable linear equations solvers* (**KSP**) is used here again with an iterative method obtained from PETSc Richardson solver (`richardson`) with block Jacobian preconditioning (`bjacobi`).
        """

        t0 = process_time()

        # Multi-lithology case
        if self.stratNb > 0:
            # Coarse sediment
            # Get erosion rate (m/yr) to volume
            self.tmpL.setArray(self.thCoarse)
            self.dm.localToGlobal(self.tmpL, self.tmp)
            self.tmp.pointwiseMult(self.tmp, self.areaGlobal)
            # Get the volume of sediment transported in m3 per year
            self._solve_KSP(False, self.fMat, self.tmp, self.vSed)
            # Update local vector
            self.dm.globalToLocal(self.vSed, self.vSedLocal)

            # Fine sediment
            if self.stratF is not None:
                # Get erosion rate (m/yr) to volume
                self.tmpL.setArray(self.thFine)
                self.dm.localToGlobal(self.tmpL, self.tmp)
                self.tmp.pointwiseMult(self.tmp, self.areaGlobal)
                # Get the volume of sediment transported in m3 per year
                self._solve_KSP(False, self.fMat, self.tmp, self.vSedf)
                # Update local vector
                self.dm.globalToLocal(self.vSedf, self.vSedfLocal)

            # Weathered sediment
            if self.stratW is not None:
                # Get erosion rate (m/yr) to volume
                self.tmpL.setArray(self.thClay)
                self.dm.localToGlobal(self.tmpL, self.tmp)
                self.tmp.pointwiseMult(self.tmp, self.areaGlobal)
                # Get the volume of sediment transported in m3 per year
                self._solve_KSP(False, self.fMat, self.tmp, self.vSedw)
                # Update local vector
                self.dm.globalToLocal(self.vSedw, self.vSedwLocal)

            # Carbonate sediment
            if self.carbOn:
                # Get erosion rate (m/yr) to volume
                self.tmpL.setArray(self.thCarb)
                self.dm.localToGlobal(self.tmpL, self.tmp)
                self.tmp.pointwiseMult(self.tmp, self.areaGlobal)
                # Get the volume of sediment transported in m3 per year
                self._solve_KSP(False, self.fMat, self.tmp, self.vSedc)
                # Update local vector
                self.dm.globalToLocal(self.vSedc, self.vSedcLocal)

        else:
            # Get erosion rate (m/yr) to volume
            self.Eb.copy(result=self.tmp)
            self.tmp.pointwiseMult(self.tmp, self.areaGlobal)
            # Get the volume of sediment transported in m3 per year
            self._solve_KSP(False, self.fMat, self.tmp, self.vSed)
            # Update local vector
            self.dm.globalToLocal(self.vSed, self.vSedLocal)

        if MPIrank == 0 and self.verbose:
            print(
                "Update Sediment Load (%0.02f seconds)" % (process_time() - t0),
                flush=True,
            )

        return

    def _matrixDir(self):
        """
        This function defines the transport direction matrix for filled-limited elevations.

        .. note::

            Each matrix is built incrementally looping through the number of flow direction paths defined by the user. It proceeds by assembling a local Compressed Sparse Row (**CSR**) matrix to a global PETSc matrix.

            When setting up transport direction matrix in PETSc, we preallocate the non-zero entries of the matrix before starting filling in the values. Using PETSc sparse matrix storage scheme has the advantage that matrix-vector multiplication is extremely fast.

        The  matrix coefficients consist of weights (comprised between 0 and 1) calculated based on the number of downslope neighbours and proportional to the slope.

        """

        dirMat = self.zMat.copy()
        indptr = np.arange(0, self.lpoints + 1, dtype=petsc4py.PETSc.IntType)
        nodes = indptr[:-1]

        for k in range(0, self.flowDir):
            # Flow direction matrix for a specific direction
            tmpMat = self._matrix_build()
            data = self.lWght[:, k].copy()
            data[self.lRcv[:, k].astype(petsc4py.PETSc.IntType) == nodes] = 0.0
            tmpMat.assemblyBegin()
            tmpMat.setValuesLocalCSR(
                indptr,
                self.lRcv[:, k].astype(petsc4py.PETSc.IntType),
                data,
            )
            tmpMat.assemblyEnd()

            # Add the weights from each direction
            dirMat += tmpMat

            tmpMat.destroy()

        del data, indptr, nodes
        gc.collect()

        # Store flow accumulation matrix
        self.dMat = dirMat.transpose().copy()

        dirMat.destroy()

        return

    def distributeSea(self, Qs):
        """
        From evaluation of enclosed seas (or endorheic basins with elevation below sea-level), this function computes the amount of incoming sediment which are deposited as well as the sediment volume able to flow downstream.

        The function calls the fortran subroutines `distsea` that takes the depressions parameters (sink IDs and overspilling node ID) to compute the different volumes and `distexcess` that moves excess sediment volumes to the overspilling nodes.

        It also calls the following *function*:

        - hierarchicalSink

        :arg Qs: numpy array of incoming sediment volume entering depressions

        :return: Qs (outflowing sediment volume numpy array)
        """

        ndepo = np.zeros(self.mpoints, dtype=np.float64)

        # Distribute sediments to next sink
        if MPIrank == 0:
            Qs, ndepo, outVol = distsea(
                Qs,
                self.gZ,
                self.sealevel,
                self.sGrp,
                self.sPits,
                self.sVol,
                self.sOut,
            )
        ndepo = MPI.COMM_WORLD.bcast(ndepo, root=0)
        if self.flatModel:
            ndepo[self.glBorders] = 0.0
        self.depo += ndepo
        self.gZ += ndepo

        # Was the elevation updated? if so update the downstream parameters
        if (ndepo > 0).any():
            self.hierarchicalSink(False, True, False, self.gZ)

        # Are there excess sea deposits to distribute?
        if MPIrank == 0:
            if (outVol > 0).any():
                Qs, ndepo = distexcess(
                    Qs,
                    self.gZ,
                    self.lFill,
                    self.lPits,
                    self.lVol,
                    outVol,
                    self.sOut,
                    self.lPits[:, 0],
                )
        Qs = MPI.COMM_WORLD.bcast(Qs, root=0)
        ndepo = MPI.COMM_WORLD.bcast(ndepo, root=0)
        if self.flatModel:
            ndepo[self.glBorders] = 0.0
            Qs[self.glBorders] = 0.0
        self.depo += ndepo
        self.gZ += ndepo

        del ndepo
        gc.collect()

        return Qs

    def distributeContinent(self, Qs, sinkFlx):
        """
        This function takes the incoming sediment volume for each continental depression and finds the volume requires to fill the sinks up to overspilling as well as the volume available for distribution in downstream regions.

        The function calls the following *function*:

        - hierarchicalSink

        :arg Qs: numpy array of incoming sediment volume entering depressions
        :arg sinkFlx: numpy array of incoming sediment volume entering open ocean

        :return: Qs, sinkFlx (updated outflowing continental and open ocean sediment volumes numpy arrays)
        """

        self.tmpL.setArray(Qs[self.locIDs])
        self.dm.localToGlobal(self.tmpL, self.tmp)

        # Move to the first downstream nodes from the overspilling ones
        self.dMat.mult(self.tmp, self.tmp1)

        # Downstream distribution of the excess in continental regions
        self.hierarchicalSink(False, False, True, self.gZ)
        self._solve_KSP(False, self.fMat, self.tmp1, self.tmp)

        self.dm.globalToLocal(self.tmp, self.tmpL)
        nQs = np.zeros(self.mpoints)
        nQs[self.locIDs] = self.tmpL.getArray().copy()
        nQs[self.outIDs] = 0.0
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, nQs, op=MPI.MAX)
        if self.flatModel:
            nQs[self.glBorders] = 0.0

        # Update open ocean incoming sediment volume and downstream continental depression ones
        Qs = np.zeros(self.mpoints)
        if sinkFlx is not None:
            sinkFlx[self.sinkID] += nQs[self.sinkID]
        nQs[self.sinkID] = 0.0
        Qs[self.lnoRcv] = nQs[self.lnoRcv]
        nQs[self.lnoRcv] = 0.0

        if sinkFlx is not None:
            # Update sediment volume available in downstream portion of the rivers
            self.tmpL.setArray(nQs[self.locIDs])
            self.vSedLocal.axpy(1.0, self.tmpL)
            self.dm.localToGlobal(self.vSedLocal, self.vSed)
        else:
            # Update water volume available in downstream portion of the rivers
            self.tmpL.setArray(nQs[self.locIDs])
            self.FAL.axpy(1.0, self.tmpL)
            self.dm.localToGlobal(self.FAL, self.FAG)

        # Are the fluxes ending in continental sinks? If so find the overspilling volume if any.
        if (Qs > 0).any():
            ndepo = np.zeros(self.mpoints, dtype=np.float64)
            if MPIrank == 0:
                ids = np.where((Qs > 0.0) & (self.lPits[:, 1] == -1))[0]
                nQs = Qs[ids]
                Qs[ids] = 0.0
                outNb, depFill = self.grpSink.sum(Qs)
                with np.errstate(divide="ignore", over="ignore"):
                    scaleV = np.divide(
                        depFill,
                        self.lVol,
                        out=np.zeros_like(self.lVol),
                        where=self.lVol != 0,
                    )
                scaleV[scaleV > 1.0] = 1.0
                ndepo = (self.lFill - self.gZ) * scaleV[self.lPits[:, 0]]
                excess = np.where(depFill > self.lVol)[0]
                Qs.fill(0.0)
                Qs[self.lOut[excess]] = depFill[excess] - self.lVol[excess]
                Qs[ids] += nQs
                del ids, outNb, depFill, scaleV, excess

            Qs = MPI.COMM_WORLD.bcast(Qs, root=0)
            ndepo = MPI.COMM_WORLD.bcast(ndepo, root=0)
            if self.flatModel:
                ndepo[self.glBorders] = 0.0
                Qs[self.glBorders] = 0.0
            self.depo += ndepo
            self.gZ += ndepo
            del ndepo

        del nQs
        gc.collect()

        if sinkFlx is not None:
            return Qs, sinkFlx
        else:
            return Qs

    def _smoothSurface(self, topo):
        """
        This function uses a diffusion approach to smooth the bathymetry before performing marine deposition. In the marine realm, we assume that local slope plays a small role on sediment distribution and the smoothed surface is used to evaluate marine multiple flow directions.

        :arg topo: numpy array of initial surface to smooth

        :return: topo (smoothed surface numpy array)
        """

        # Diffusion matrix construction
        Cd = np.full(self.lpoints, 1.0e-4, dtype=np.float64)
        seaID = np.where(topo[self.locIDs] < self.sealevel)
        Cd[seaID] = 1.0e5  # this is an hardcoded value for now...

        diffCoeffs = sethillslopecoeff(self.lpoints, Cd * self.dt)
        diff = self._matrix_build_diag(diffCoeffs[:, 0])
        indptr = np.arange(0, self.lpoints + 1, dtype=petsc4py.PETSc.IntType)

        for k in range(0, self.maxnb):
            tmpMat = self._matrix_build()
            indices = self.FVmesh_ngbID[:, k].copy()
            data = diffCoeffs[:, k + 1]
            ids = np.nonzero(data == 0.0)
            indices[ids] = ids
            tmpMat.assemblyBegin()
            tmpMat.setValuesLocalCSR(
                indptr,
                indices.astype(petsc4py.PETSc.IntType),
                data,
            )
            tmpMat.assemblyEnd()
            diff += tmpMat
            tmpMat.destroy()

        # Get smoothed elevation values
        self.tmp1.setArray(topo[self.glbIDs])
        self._solve_KSP(False, diff, self.tmp1, self.tmp)
        self.dm.globalToLocal(self.tmp, self.tmpL)
        hl = self.tmpL.getArray().copy()
        topo.fill(0.0)
        topo[self.locIDs] = hl
        topo[self.outIDs] = -1.0e8
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, topo, op=MPI.MAX)

        diff.destroy()

        del ids, indices, indptr, data, diffCoeffs, Cd, seaID, hl
        gc.collect()

        return topo

    def _marineDeposition(self, h, stype):
        """
        This function computes the maximum elevations of marine deposition based on clinoform slopes and antecedent shelf and continental slope physiographies.

        It also calls the following *private function*:

        - _smoothSurface

        :arg h: numpy array of local elevation values
        :arg stype: sediment type (integer)

        :return: smthZ, hclino (smoothed surface and maximum marine deposition elevations numpy arrays)
        """

        # Marine smoothed bathymetry
        smthZ = np.zeros(self.mpoints)
        smthZ[self.locIDs] = h
        smthZ[self.outIDs] = -1.0e8
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, smthZ, op=MPI.MAX)
        smthZ = self._smoothSurface(smthZ)

        # From the distance to coastline define the upper limit of the shelf to ensure a maximum slope angle
        clinoH = self.sealevel - 1.0 - self.coastDist * self.slps[stype] * 1.0e-5

        # Get the maximum marine deposition heights
        depo = smthZ[self.locIDs] + 500.0 - h
        shelfIDs = smthZ[self.locIDs] + 500.0 > clinoH
        depo[shelfIDs] = clinoH[shelfIDs] - h[shelfIDs]
        depo[depo < 0.0] = 0.0

        # From the marine deposition heights define the marine deposition elevations
        hclino = np.zeros(self.mpoints, dtype=np.float64) + 1.0e8
        hclino[self.locIDs] = h + depo
        hclino[self.outIDs] = 1.0e8
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, hclino, op=MPI.MIN)

        del depo, clinoH, shelfIDs
        gc.collect()

        return smthZ, hclino

    def _diffuseOcean(self, zb, ndepo, stype):
        """
        For sediment reaching the marine realm, this function computes the related marine deposition diffusion. The approach is based on a linear diffusion.

        The diffusion equation is ran for the coarser sediment first and for the finest ones afterwards. This mimicks the standard behaviour observed in stratigraphic architectures where the fine fraction are generally transported over longer distances.

        :arg zb: numpy array of current elevation over which sediment diffusion is performed
        :arg ndepo: numpy array of incoming marine depositional thicknesses
        :arg stype: sediment type (integer)

        :return: ndepo, h (updated elevations and diffused depositions numpy arrays)
        """

        # Get diffusion coefficients based on sediment type
        sedK = self.sedimentK
        if sedK > 0.0:
            if stype == 1:
                sedK = self.sedimentKf
            if stype == 3:
                sedK = self.sedimentKw

        # Top elevations to diffuse
        h = zb + ndepo

        # Diffusion matrix construction
        Cd = np.zeros(self.lpoints, dtype=np.float64)
        Cd[self.seaID] = sedK
        scale = np.divide(ndepo[self.locIDs], ndepo[self.locIDs] + 500.0)
        Cd *= scale

        diffCoeffs = sethillslopecoeff(self.lpoints, Cd * self.dt)
        seaDiff = self._matrix_build_diag(diffCoeffs[:, 0])
        indptr = np.arange(0, self.lpoints + 1, dtype=petsc4py.PETSc.IntType)
        del Cd, scale

        for k in range(0, self.maxnb):
            tmpMat = self._matrix_build()
            indices = self.FVmesh_ngbID[:, k].copy()
            data = diffCoeffs[:, k + 1]
            ids = np.nonzero(data == 0.0)
            indices[ids] = ids
            tmpMat.assemblyBegin()
            tmpMat.setValuesLocalCSR(
                indptr,
                indices.astype(petsc4py.PETSc.IntType),
                data,
            )
            tmpMat.assemblyEnd()
            seaDiff += tmpMat
            tmpMat.destroy()

        del ids, data, indices, indptr, diffCoeffs
        gc.collect()

        # Perform diffusion equation
        self.tmp1.setArray(h[self.glbIDs])
        self._solve_KSP(False, seaDiff, self.tmp1, self.tmp)
        self.dm.globalToLocal(self.tmp, self.tmpL)
        hl = self.tmpL.getArray().copy()
        h = np.zeros(self.mpoints)
        h[self.locIDs] = hl
        h[self.outIDs] = -1.0e8
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, h, op=MPI.MAX)
        ndepo = h - zb
        ndepo[ndepo < 0.0] = 0.0

        seaDiff.destroy()
        if self.flatModel:
            h[self.glBorders] = zb[self.glBorders]

        del hl
        gc.collect()

        return ndepo, h

    def _distributeOcean(self, sinkFlx, stype=0):
        """
        Main entry point to distribute incoming river sediment fluxes in the open ocean and perform diffusion of freshly deposited accumulations.

        The function calls the following *private functions*:

        - _marineDeposition
        - _diffuseOcean

        The diffusion equation is ran for the coarser sediment first and for the finest ones afterwards. This mimicks the standard behaviour observed in stratigraphic architectures where the fine fraction are generally transported over longer distances.

        :arg sinkFlx: numpy array of incoming sediment volume entering open ocean
        :arg stype: sediment type (integer)
        """

        exp = 1.0e-2
        flowDir = 8
        zb = self.gZ.copy()
        smthZ, hclino = self._marineDeposition(zb[self.locIDs], stype)

        maxDist = 2.0e6
        coastDist = np.zeros(self.mpoints, dtype=np.float64)
        coastDist[self.locIDs] = self.coastDist
        coastDist[self.outIDs] = 0.0
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, coastDist, op=MPI.MAX)
        ins = np.where(coastDist > maxDist)[0]
        if self.flatModel:
            ins = self.glBorders.copy()
        gZ = self.gZ.copy()
        smthZ[ins] = -3.0e4
        zsmth = smthZ.copy()
        sinkFlx = sinkFlx / float(self.diffstep)

        ndep = np.zeros(self.mpoints, dtype=np.float64)
        ndepo = np.zeros(self.mpoints, dtype=np.float64)
        for step in range(self.diffstep):
            ndepo.fill(0.0)
            sRcvs, sWghts = mfdrcvs(
                flowDir, exp, self.inIDs, smthZ[self.locIDs], -2.9e4
            )
            if self.flatModel:
                sRcvs[self.idBorders, :] = np.tile(self.idBorders, (flowDir, 1)).T
                sWghts[self.idBorders, :] = 0.0

            # Distribute open sea sediments
            sWght = np.zeros((self.mpoints, flowDir), dtype=np.float64)
            sWght[self.locIDs, :] = sWghts
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, sWght, op=MPI.MAX)

            sRcv = -np.ones((self.mpoints, flowDir), dtype=int)
            val = self.locIDs[sRcvs]
            ins = np.where(sRcvs == -1)[0]
            val[ins] = -1
            sRcv[self.locIDs, :] = val
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, sRcv, op=MPI.MAX)

            if MPIrank == 0:
                sid = np.argsort(smthZ)[::-1]
                topID = np.where(sinkFlx[sid] > 0)[0][0]
                vdepth = hclino - gZ
                vdepth[vdepth < 0.0] = 0.0
                ndepo = distocean(
                    flowDir, sid[topID:], sinkFlx, sRcv, sWght, self.marea, vdepth
                )
            ndepo = MPI.COMM_WORLD.bcast(ndepo, root=0)
            if self.flatModel:
                ndepo[self.glBorders] = 0.0
            ndep += ndepo
            ndep, gZ = self._diffuseOcean(zb, ndep, stype)
            if self.flatModel:
                ndep[self.glBorders] = 0.0
            smthZ = zsmth + ndep

        self.depo += ndep  # gZ - self.oldh
        self.gZ = gZ.copy()

        del hclino, coastDist, gZ, smthZ, zsmth
        del sinkFlx, ndep, ndepo, sRcvs, sWghts, val
        if MPIrank == 0:
            del sid, vdepth
        gc.collect()

        return

    def _currentElevation(self):
        """
        Get current elevation.
        """

        hl = self.hLocal.getArray().copy()
        gZ = np.zeros(self.mpoints)
        gZ[self.locIDs] = hl
        gZ[self.outIDs] = -1.0e8
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, gZ, op=MPI.MAX)
        self.gZ = gZ.copy()
        self.oldh = gZ.copy()

        del hl
        gc.collect()

        return gZ

    def _fill2D(self, minz, elev, fmax=1.0e6):
        """
        Fill pit for the 2D case.
        """

        nelev = elev.copy()
        nelev[self.glBorders] = minz - 1.0
        fill, pits = fillpit(minz, nelev, fmax)
        fill[self.glBorders] = elev[self.glBorders]
        pits[self.glBorders] = -1

        del nelev
        gc.collect()

        return fill, pits

    def _nullitfyBorders(self, rcv, wght):
        """
        Set receivers and weights to zero for border points when running a 2D case.
        """

        rcv[self.idBorders, :] = np.tile(self.idBorders, (self.flowDir, 1)).T
        wght[self.idBorders, :] = 0.0

        return rcv, wght

    def _getDepressions(self):
        """
        Get continental depressions parameters.
        """

        sum_weight = np.sum(self.lWght, axis=1)
        tmpw = np.zeros(self.mpoints, dtype=np.float64)
        tmpw[self.locIDs] = sum_weight
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, tmpw, op=MPI.MAX)
        self.lnoRcv = tmpw == 0.0
        del tmpw

        if MPIrank == 0:
            # Compute continental pit volumes
            self.grpSink = npi.group_by(self.lPits[:, 0])
            self.lGrp = self.grpSink.unique
            pitNb, self.lVol = self.grpSink.sum((self.lFill - self.gZ) * self.marea)
            _, outids, _ = np.intersect1d(self.lPits[:, 0], pitNb, return_indices=True)
            self.lOut = self.lPits[outids, 1]
            del pitNb, outids

        return

    def _getCloseSea(self):
        """
        Get closed sea parameters.
        """

        # Set all nodes below sea-level as sinks
        self.sinkID = np.where(self.sFill < self.sealevel)[0]

        if MPIrank == 0:
            # Find closed seas
            closeIDs = np.where(
                (self.sFill > self.sealevel) & (self.gZ < self.sealevel)
            )[0]
            grpSink = npi.group_by(self.sPits[closeIDs, 0])
            self.sGrp = grpSink.unique
            self.sIns, self.sOut, self.sVol = seaparams(
                self.gZ, self.sealevel, self.sGrp, self.sPits
            )
            del closeIDs, grpSink

        return

    def hierarchicalSink(self, offshore=False, land=False, limited=False, gZ=None):
        """
        This function defines the depression properties used to evaluate downstream sediment transport. To do so, the following three surfaces could be considered:

        1. a completely filled surface starting from the deep offshore regions
        2. a completely filled surface starting from the sea-level position
        3. a filled-limited surface starting from the sea-level position and limited by user defined value (to distribute sediment flux)

        In addition, we extract the volume of all depressions based on current elevation, depressionless one and voronoi cell areas. It also stores the spillover vertices indices for each of the depression. It is ran over the global mesh.

        .. note::

            This function uses the **numpy-indexed** library which contains functionality for indexed operations on numpy ndarrays and provides efficient vectorized functionality such as grouping and set operations.

        :arg offshore: boolean to compute sinks and associated parameters for surface 1
        :arg land: boolean to compute sinks and associated parameters for surface 2
        :arg limited: boolean to compute sinks and associated parameters for surface 3
        """

        # Get current state global elevations...
        if gZ is None:
            gZ = self._currentElevation()

        mingZ = -4.0e3
        if self.flatModel:
            mingZ = gZ.min()

        # Perform pit filling on process rank 0
        if MPIrank == 0:
            if offshore:
                if self.flatModel:
                    sFill, self.sPits = self._fill2D(mingZ, gZ, 1.0e6)
                else:
                    sFill, self.sPits = fillpit(-4.0e3, gZ, 1.0e6)
            if land:
                if self.flatModel:
                    lFill, self.lPits = self._fill2D(self.sealevel, gZ, 1.0e6)
                else:
                    lFill, self.lPits = fillpit(self.sealevel, gZ, 1.0e6)
            if limited:
                fill = self.sedfill
                if self.flowDist:
                    fill = self.waterfill
                if self.flatModel:
                    lFill, self.lPits = self._fill2D(self.sealevel, gZ, fill)
                else:
                    lFill, self.lPits = fillpit(self.sealevel, gZ, fill)
        else:
            if offshore:
                sFill = np.zeros(self.mpoints, dtype=np.float64)
            if land or limited:
                lFill = np.zeros(self.mpoints, dtype=np.float64)

        if offshore:
            self.sFill = MPI.COMM_WORLD.bcast(sFill, root=0)
            self._getCloseSea()

            # Define multiple flow directions for filled elevation
            self.sRcv, _, self.sWght = mfdreceivers(
                self.flowDir,
                self.flowExp,
                self.inIDs,
                self.sFill[self.locIDs],
                mingZ,
            )
            if self.flatModel:
                self.sRcv, self.sWght = self._nullitfyBorders(self.sRcv, self.sWght)
            del sFill

        if land or limited:
            self.lFill = MPI.COMM_WORLD.bcast(lFill, root=0)
            del lFill
            # Define multiple flow directions for filled elevation
            self.lRcv, _, self.lWght = mfdreceivers(
                self.flowDir,
                self.flowExp,
                self.inIDs,
                self.lFill[self.locIDs],
                self.sealevel,
            )
            if self.flatModel:
                self.lRcv, self.lWght = self._nullitfyBorders(self.lRcv, self.lWght)
            if land:
                self._matrixDir()
            else:
                self.matrixFlow(False)
            # Get depression parameters
            self._getDepressions()

        gc.collect()

        return

    def _getSedVol(self, stype):
        """
        Pick the relevant PETSc array for the specified sediment type.
        """

        if stype == 0:
            self.vSedLocal.copy(result=self.QsL)
        elif stype == 1:
            self.vSedfLocal.copy(result=self.QsL)
        elif stype == 2:
            self.vSedcLocal.copy(result=self.QsL)
        elif stype == 3:
            self.vSedwLocal.copy(result=self.QsL)
        self.dm.localToGlobal(self.QsL, self.Qs)

        return

    def _distributeSediment(self, stype=0):
        """
        Main entry point to distribute rivers' sediment fluxes over the surface. It accounts for both continental and marine deposition based on previously eroded sediments from rivers. The algorithm conserves sediment volumes and accounts for downstream deposition in both continental sink (endorheic basins), enclosed sea or lakes and open ocean environments.

        .. note::

            The deposition in depressions (both marine and continental) is performed based on the differences between sink volumes and incoming sediments ones and scale deposition accordingly. For the open seas, deposited sediment thicknesses are defined based on clinorform slopes and shelf and continental slopes physiographies before being diffused based on user diffusion coefficient.

        The function calls the following *private and non-private functions*:

        - hierarchicalSink
        - _getSedVol
        - distributeSea
        - distributeContinent
        - _distributeOcean

        :arg stype: sediment type (integer)
        """

        t0 = process_time()
        # Get downstream sinks in marine and continental regions
        self.hierarchicalSink(True, True, False, None)

        # Define the volume to distribute
        self._getSedVol(stype)

        # Get the volumetric sediment rate (m3/yr) to distribute during the time step and convert in volume (m3) for considered timestep
        lQs = self.QsL.getArray().copy()
        nQs = np.zeros(self.mpoints, dtype=np.float64)
        nQs[self.locIDs] = lQs * self.dt
        nQs[self.outIDs] = 0.0
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, nQs, op=MPI.MAX)
        Qs = np.zeros(self.mpoints, dtype=np.float64)
        if self.flatModel:
            Qs[self.glBorders] = 0.0
        Qs[self.pitPts] = nQs[self.pitPts]
        del nQs, lQs

        # Update all flux in the main sink
        sinkFlx = np.zeros(self.mpoints, dtype=np.float64)
        self.depo = np.zeros(self.mpoints, dtype=np.float64)
        sinkFlx[self.sinkID] += Qs[self.sinkID]
        Qs[self.sinkID] = 0.0

        step = 0
        while (Qs > 0).any():

            # Check if sediments are in closed sea
            inQs = np.where(Qs > 0)[0]
            insea = np.zeros(1, dtype=np.int64)
            if MPIrank == 0:
                if (self.sIns[inQs] > -1).any():
                    insea[0] = 1
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, insea, op=MPI.MAX)
            del inQs
            if insea[0] > 0:
                # Distribute sediments in closed sea
                Qs = self.distributeSea(Qs)
                # Remove flux that flow into open sea
                sinkFlx[self.sinkID] += Qs[self.sinkID]
                Qs[self.sinkID] = 0.0
                if self.flatModel:
                    Qs[self.glBorders] = 0.0

            # At this point excess sediments are all in continental regions
            if (Qs > 0).any():
                Qs, sinkFlx = self.distributeContinent(Qs, sinkFlx)
                if self.flatModel:
                    Qs[self.glBorders] = 0.0

            step += 1
            if step > 20:
                if MPIrank == 0:
                    print("Forced Distribution to Ocean")
                sinkFlx += Qs
                break
        del Qs
        if MPIrank == 0 and self.verbose:
            print("Distribution step nb:", step)
            print(
                "Distribute Continental Sediments (%0.02f seconds)"
                % (process_time() - t0),
                flush=True,
            )

        # Distribute ocean sediments
        if (sinkFlx > 0).any():
            t0 = process_time()
            if self.flatModel:
                sinkFlx[self.glBorders] = 0.0
            self._distributeOcean(sinkFlx, stype)
            if MPIrank == 0 and self.verbose:
                print(
                    "Distribute Ocean Sediments (%0.02f seconds)"
                    % (process_time() - t0),
                    flush=True,
                )

        del sinkFlx
        gc.collect()

        # Update cumed and elev
        self.tmpL.setArray(self.depo[self.locIDs])
        self.dm.localToGlobal(self.tmpL, self.tmp)
        self.cumED.axpy(1.0, self.tmp)
        self.hGlobal.axpy(1.0, self.tmp)
        self.dm.globalToLocal(self.cumED, self.cumEDLocal)
        self.dm.globalToLocal(self.hGlobal, self.hLocal)

        # Update stratigraphic layer parameters
        if self.stratNb > 0:
            self.deposeStrat(stype)
            self.elevStrat()

        return

    def sedChange(self):
        """
        This function is the main entry point to perform both continental and marine river-induced deposition. It calls the private function `_distributeSediment`.
        """

        if self.stratNb > 0:
            self._distributeSediment(stype=0)
            if self.carbOn:
                self._distributeSediment(stype=2)
            if self.stratF is not None:
                self._distributeSediment(stype=1)
            if self.stratW is not None:
                self._distributeSediment(stype=3)
        else:
            self._distributeSediment(stype=0)

        return

    def getHillslope(self):
        r"""
        This function computes hillslope processes. The code assumes that gravity is the main driver for transport and states that the flux of sediment is proportional to the gradient of topography.

        As a result, we use a linear diffusion law commonly referred to as **soil creep**:

        .. math::
          \frac{\partial z}{\partial t}= \kappa_{D} \nabla^2 z

        in which :math:`\kappa_{D}` is the diffusion coefficient and can be defined with different values for the marine and land environments (set with `hillslopeKa` and `hillslopeKm` in the YAML input file). It encapsulates, in a simple formulation, processes operating on superficial sedimentary layers. Main controls on variations of :math:`\kappa_{D}` include substrate, lithology, soil depth, climate and biological activity.
        """

        # Compute Hillslope Diffusion Law
        h = self.hLocal.getArray().copy()
        self.seaID = np.where(h <= self.sealevel)[0]
        self._hillSlope()

        # Update layer elevation
        if self.stratNb > 0:
            self.elevStrat()

        del h
        gc.collect()

        return

    def _growCarbonates(self):
        """
        When carbonates is turned on update carbonate thicknesses based on fuzzy logic controls. The carbonate thicknesses created are uncompacted ones.

        .. warning::

            This function is a place order and will be updated in a future version of `gospl`.
        """

        # Limit fuzzy controllers to possible value range
        hl = self.sealevel - self.hLocal.getArray().copy()

        # TO DO: this will need to be updated with additional controls...
        carbH = np.zeros(self.lpoints, dtype=np.float64)
        validIDs = np.ones(self.lpoints, dtype=np.int64)
        for k in range(len(self.carbCtrl.controlKeys)):
            # Accomodation control
            if self.carbCtrl.controlKeys[k] == "depth":
                ids = np.where(
                    np.logical_or(
                        hl <= self.carbCtrl.controlBounds[k][0],
                        hl >= self.carbCtrl.controlBounds[k][1],
                    )
                )[0]
                validIDs[ids] = 0

        # Compute carbonate growth rates in metres per thousand years
        # and associated thicknesses
        ids = np.where(validIDs > 0)[0]
        for k in range(len(ids)):
            self.carbCtrl.carbControlSystem.input["depth"] = hl[ids[k]]
            # self.carbCtrl.carbControlSystem.input["temperature"] = 20.0
            self.carbCtrl.carbControlSystem.compute()
            growthRate = self.carbCtrl.carbControlSystem.output["growth"]
            carbH[ids[k]] = min(hl[ids[k]] - 1.0e-6, growthRate * self.dt * 1.0e-3)

        # Update top layers based on carbonate growth
        self.tmpL.setArray(carbH)
        self.dm.localToGlobal(self.tmpL, self.tmp)
        self.cumED.axpy(1.0, self.tmp)
        self.hGlobal.axpy(1.0, self.tmp)
        self.dm.globalToLocal(self.cumED, self.cumEDLocal)
        self.dm.globalToLocal(self.hGlobal, self.hLocal)

        if self.stratNb > 0:
            # Update carbonate reef content in the stratigraphic layer
            self.deposeStrat(2)

        del ids, validIDs, carbH, hl
        gc.collect()

        return

    def _hillSlope(self):
        r"""
        This function computes hillslope using a linear diffusion law commonly referred to as **soil creep**:

        .. math::
          \frac{\partial z}{\partial t}= \kappa_{D} \nabla^2 z

        in which :math:`\kappa_{D}` is the diffusion coefficient and can be defined with different values for the marine and land environments (set with `hillslopeKa` and `hillslopeKm` in the YAML input file).

        .. note::
            The hillslope processes in `gospl` are considered to be happening at the same rate for coarse and fine sediment sizes.

        """

        if self.Cda == 0.0 and self.Cdm == 0.0:
            return

        t0 = process_time()

        # Diffusion matrix construction
        Cd = np.full(self.lpoints, self.Cda, dtype=np.float64)
        Cd[self.seaID] = self.Cdm

        diffCoeffs = sethillslopecoeff(self.lpoints, Cd * self.dt)
        diffMat = self._matrix_build_diag(diffCoeffs[:, 0])
        indptr = np.arange(0, self.lpoints + 1, dtype=petsc4py.PETSc.IntType)

        for k in range(0, self.maxnb):
            tmpMat = self._matrix_build()
            indices = self.FVmesh_ngbID[:, k].copy()
            data = diffCoeffs[:, k + 1]
            ids = np.nonzero(data == 0.0)
            indices[ids] = ids
            tmpMat.assemblyBegin()
            tmpMat.setValuesLocalCSR(
                indptr,
                indices.astype(petsc4py.PETSc.IntType),
                data,
            )
            tmpMat.assemblyEnd()
            diffMat += tmpMat
            tmpMat.destroy()
        del ids, indices, indptr, diffCoeffs, Cd
        gc.collect()

        # Get elevation values for considered time step
        self.hGlobal.copy(result=self.hOld)
        self._solve_KSP(True, diffMat, self.hOld, self.hGlobal)
        diffMat.destroy()

        # Update cumulative erosion/deposition and elevation
        self.tmp.waxpy(-1.0, self.hOld, self.hGlobal)
        self.cumED.axpy(1.0, self.tmp)
        self.dm.globalToLocal(self.cumED, self.cumEDLocal)
        self.dm.globalToLocal(self.hGlobal, self.hLocal)

        if self.stratNb > 0:
            self.erodeStrat()
            if self.stratW is not None:
                self.deposeStrat(3)
            else:
                self.deposeStrat(0)

        if MPIrank == 0 and self.verbose:
            print(
                "Compute Hillslope Processes (%0.02f seconds)" % (process_time() - t0),
                flush=True,
            )

        return



##############################
# BUILDING CONDA ENVIRONMENT
##############################
#
# conda env export --from-history | grep -v "prefix" > conda-env.yml
# conda env create -f conda-env.yml
# conda activate gospl-package
#
##############################
# BUILDING PYTHON PACKAGE PYPI
##############################
#
# python3 -m pip install --user --upgrade setuptools wheel
# python3 setup.py sdist
# python3 -m pip install --user --upgrade twine
# /usr/local/bin/twine check dist/*
# /usr/local/bin/twine upload dist/*
#
##############################

import os
import io
import sys
import subprocess
from setuptools import find_packages

try:
    subprocess.call([sys.executable, "-m", "pip", "install", "numpy"])
except ImportError:
    pass

from numpy.distutils.core import setup, Extension
from distutils.command.sdist import sdist

try:
    from distutils.command import bdist_conda
except ImportError:
    pass

# in development set version to none and ...
PYPI_VERSION = "0.2.6"

# Place install_requires into the text file "requirements.txt"
with open("requirements.txt") as f2:
    requirements = f2.read().strip().splitlines()


# class sdist(_sdist):
#     def run(self):
#         _sdist.run(self)


packs = find_packages(include=["gospl", "gospl.*"])


def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(["git", "rev-parse", "--short", "HEAD"])
        GIT_REVISION = out.strip().decode("ascii")
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


if PYPI_VERSION is None:
    PYPI_VERSION = git_version()


ext = Extension(
    name="gospl._fortran", sources=["fortran/functions.pyf", "fortran/functions.F90"]
)


this_directory = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(this_directory, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()


if __name__ == "__main__":
    setup(
        name="gospl",
        author="Tristan Salles",
        author_email="tristan.salles@sydney.edu.au",
        url="https://github.com/Geodels/gospl",
        version=PYPI_VERSION,
        license="GPLv3",
        description="A Python interface to perform Global Landscape Evolution Model",
        keywords=[
            "python",
            "paleogeography",
            "sediment-transport",
            "paleoclimate",
            "model",
            "landscape",
            "landscape-evolution",
            "basin-modeling",
            "erosion-process",
            "compaction",
            "lithology",
            "science",
        ],
        long_description=long_description,
        long_description_content_type="text/markdown",
        ext_modules=[ext],
        # packages=["gospl", "gospl.tools", "gospl.flow", "gospl.mesher", "gospl.sed"],
        packages=packs,
        install_requires=requirements,
        setup_requires=[[p for p in requirements if p.startswith("numpy")][0]],
        python_requires=">=3",
        classifiers=[
            "Intended Audience :: Science/Research",
            "Programming Language :: Fortran",
            "Operating System :: Unix",
            "Operating System :: MacOS",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],
        cmdclass={"sdist": sdist},
    )



import os
import gc
import sys
import petsc4py
import numpy as np

from mpi4py import MPI
from time import process_time

if "READTHEDOCS" not in os.environ:
    from gospl._fortran import stratasimple
    from gospl._fortran import stratabuild
    from gospl._fortran import stratabuildcarb

petsc4py.init(sys.argv)
MPIrank = petsc4py.PETSc.COMM_WORLD.Get_rank()
MPIsize = petsc4py.PETSc.COMM_WORLD.Get_size()
MPIcomm = petsc4py.PETSc.COMM_WORLD


class STRAMesh(object):
    """
    This class encapsulates all the functions related to underlying stratigraphic information. As mentionned previously, `gospl` has the ability to track different types of clastic sediment size and one type of carbonate (still under development). Sediment compaction in stratigraphic layers geometry and properties change are also considered.

    """

    def __init__(self):
        """
        The initialisation of `STRAMesh` class related to stratigraphic informations.
        """

        self.stratH = None
        self.stratF = None
        self.stratW = None
        self.stratZ = None
        self.stratC = None

        self.phiS = None
        self.phiF = None
        self.phiC = None

        return

    def readStratLayers(self):
        """
        When stratigraphic layers are turned on, this function reads any initial stratigraphic layers provided by within the YAML input file (key: `npstrata`).

        The following variables will be read from the file:

        - thickness of each stratigrapic layer `strataH` accounting for both erosion & deposition events.
        - proportion of fine sediment `strataF` contains in each stratigraphic layer.
        - proportion of weathered sediment `strataW` contains in each stratigraphic layer.
        - elevation at time of deposition, considered to be to the current elevation for the top stratigraphic layer `strataZ`.
        - porosity of coarse sediment `phiS` in each stratigraphic layer computed at center of each layer.
        - porosity of fine sediment `phiF` in each stratigraphic layer computed at center of each layer.
        - porosity of weathered sediment `phiW` in each stratigraphic layer computed at center of each layer.
        - proportion of carbonate sediment `strataC` contains in each stratigraphic layer if the carbonate module is turned on.
        - porosity of carbonate sediment `phiC` in each stratigraphic layer computed at center of each layer when the carbonate module is turned on.

        """

        if self.strataFile is not None:
            fileData = np.load(self.strataFile)
            stratVal = fileData["strataH"]
            self.initLay = stratVal.shape[1]
            self.stratNb += self.initLay

            # Create stratigraphic arrays
            self.stratH = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.stratH[:, 0 : self.initLay] = stratVal[self.locIDs, 0 : self.initLay]

            stratVal = fileData["strataF"]
            self.stratF = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.stratF[:, 0 : self.initLay] = stratVal[self.locIDs, 0 : self.initLay]

            stratVal = fileData["strataW"]
            self.stratW = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.stratW[:, 0 : self.initLay] = stratVal[self.locIDs, 0 : self.initLay]

            stratVal = fileData["strataZ"]
            self.stratZ = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.stratZ[:, 0 : self.initLay] = stratVal[self.locIDs, 0 : self.initLay]

            stratVal = fileData["phiS"]
            self.phiS = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.phiS[:, 0 : self.initLay] = stratVal[self.locIDs, 0 : self.initLay]

            stratVal = fileData["phiF"]
            self.phiF = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.phiF[:, 0 : self.initLay] = stratVal[self.locIDs, 0 : self.initLay]

            stratVal = fileData["phiW"]
            self.phiW = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.phiW[:, 0 : self.initLay] = stratVal[self.locIDs, 0 : self.initLay]

            if self.carbOn:
                stratVal = fileData["strataC"]
                self.stratC = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
                self.stratC[:, 0 : self.initLay] = stratVal[
                    self.locIDs, 0 : self.initLay
                ]

                stratVal = fileData["phiC"]
                self.phiC = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
                self.phiC[:, 0 : self.initLay] = stratVal[self.locIDs, 0 : self.initLay]

            del fileData, stratVal
            gc.collect()
        else:
            self.stratH = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.phiS = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)
            self.stratZ = np.zeros((self.lpoints, self.stratNb), dtype=np.float64)

        return

    def deposeStrat(self, stype):
        """
        Add deposition on top of an existing stratigraphic layer. The following variables will be recorded:

        - thickness of each stratigrapic layer `stratH` accounting for both erosion & deposition events.
        - proportion of fine sediment `stratF` contains in each stratigraphic layer.
        - proportion of weathered sediment `stratW` contains in each stratigraphic layer.
        - porosity of coarse sediment `phiS` in each stratigraphic layer computed at center of each layer.
        - porosity of fine sediment `phiF` in each stratigraphic layer computed at center of each layer.
        - porosity of weathered sediment `phiW` in each stratigraphic layer computed at center of each layer.
        - proportion of carbonate sediment `stratC` contains in each stratigraphic layer if the carbonate module is turned on.
        - porosity of carbonate sediment `phiC` in each stratigraphic layer computed at center of each layer when the carbonate module is turned on.

        :arg stype: sediment type (integer)
        """

        self.dm.globalToLocal(self.tmp, self.tmpL)
        depo = self.tmpL.getArray().copy()
        depo[depo < 1.0e-4] = 0.0
        if self.stratF is not None:
            fineH = self.stratH[:, self.stratStep] * self.stratF[:, self.stratStep]
        if self.stratW is not None:
            clayH = self.stratH[:, self.stratStep] * self.stratW[:, self.stratStep]
        if self.carbOn:
            carbH = self.stratH[:, self.stratStep] * self.stratC[:, self.stratStep]
        self.stratH[:, self.stratStep] += depo
        ids = np.where(depo > 0)[0]

        if stype == 0:
            self.phiS[ids, self.stratStep] = self.phi0s
        elif stype == 1:
            fineH[ids] += depo[ids]
            self.phiF[ids, self.stratStep] = self.phi0f
        elif stype == 2:
            carbH[ids] += depo[ids]
            self.phiC[ids, self.stratStep] = self.phi0c
        elif stype == 3:
            clayH[ids] += depo[ids]
            self.phiW[ids, self.stratStep] = self.phi0w

        if self.stratF is not None:
            self.stratF[ids, self.stratStep] = (
                fineH[ids] / self.stratH[ids, self.stratStep]
            )
            del fineH
        if self.stratW is not None:
            self.stratW[ids, self.stratStep] = (
                clayH[ids] / self.stratH[ids, self.stratStep]
            )
            del clayH
        if self.carbOn:
            self.stratC[ids, self.stratStep] = (
                carbH[ids] / self.stratH[ids, self.stratStep]
            )
            del carbH

        # Cleaning arrays
        del depo, ids
        gc.collect()

        return

    def erodeStrat(self):
        """
        This function removes eroded sediment thicknesses from the stratigraphic pile. The function takes into account the porosity values of considered lithologies in each eroded stratigraphic layers.

        It follows the following assumptions:

        - Eroded thicknesses from stream power law and hillslope diffusion are considered to encompass both the solid and void phase.
        - Only the solid phase will be moved dowstream by surface processes.
        - The corresponding deposit thicknesses for those freshly eroded sediments correspond to uncompacted thicknesses based on the porosity at surface given from the input file.
        """

        self.dm.globalToLocal(self.tmp, self.tmpL)
        ero = self.tmpL.getArray().copy()
        ero[ero > 0] = 0.0

        # Nodes experiencing erosion
        nids = np.where(ero < 0)[0]
        if len(nids) == 0:
            del ero, nids
            gc.collect()
            return

        # Cumulative thickness for each node
        self.stratH[nids, 0] += 1.0e6
        cumThick = np.cumsum(self.stratH[nids, self.stratStep :: -1], axis=1)[:, ::-1]
        boolMask = cumThick < -ero[nids].reshape((len(nids), 1))
        mask = boolMask.astype(int)

        if self.stratF is not None:
            # Get fine sediment eroded from river incision
            thickF = (
                self.stratH[nids, 0 : self.stratStep + 1]
                * self.stratF[nids, 0 : self.stratStep + 1]
            )
            # From fine thickness extract the solid phase that is eroded
            thFine = thickF * (1.0 - self.phiF[nids, 0 : self.stratStep + 1])
            thFine = np.sum((thFine * mask), axis=1)

        if self.stratW is not None:
            # Get weathered sediment eroded from river incision
            thickW = (
                self.stratH[nids, 0 : self.stratStep + 1]
                * self.stratW[nids, 0 : self.stratStep + 1]
            )
            # From weathered thickness extract the solid phase that is eroded
            thClay = thickW * (1.0 - self.phiW[nids, 0 : self.stratStep + 1])
            thClay = np.sum((thClay * mask), axis=1)

        # Get carbonate sediment eroded from river incision
        if self.carbOn:
            thickC = (
                self.stratH[nids, 0 : self.stratStep + 1]
                * self.stratC[nids, 0 : self.stratStep + 1]
            )
            # From carbonate thickness extract the solid phase that is eroded
            thCarb = thickC * (1.0 - self.phiC[nids, 0 : self.stratStep + 1])
            thCarb = np.sum((thCarb * mask), axis=1)
            # From sand thickness extract the solid phase that is eroded
            thickS = (
                self.stratH[nids, 0 : self.stratStep + 1] - thickF - thickC - thickW
            )
            thCoarse = thickS * (1.0 - self.phiS[nids, 0 : self.stratStep + 1])
            thCoarse = np.sum((thCoarse * mask), axis=1)
        else:
            # From sand thickness extract the solid phase that is eroded
            if self.stratF is not None:
                thickS = self.stratH[nids, 0 : self.stratStep + 1] - thickF - thickW
            else:
                thickS = self.stratH[nids, 0 : self.stratStep + 1]
            thCoarse = thickS * (1.0 - self.phiS[nids, 0 : self.stratStep + 1])
            thCoarse = np.sum((thCoarse * mask), axis=1)

        # Clear all stratigraphy points which are eroded
        cumThick[boolMask] = 0.0
        tmp = self.stratH[nids, : self.stratStep + 1]
        tmp[boolMask] = 0
        self.stratH[nids, : self.stratStep + 1] = tmp

        # Erode remaining stratal layers
        # Get non-zero top layer number
        eroLayNb = np.bincount(np.nonzero(cumThick)[0]) - 1
        eroVal = cumThick[np.arange(len(nids)), eroLayNb] + ero[nids]

        if self.stratF is not None:
            # Get thickness of each sediment type eroded in the remaining layer
            self.thFine = np.zeros(self.lpoints)
            # From fine thickness extract the solid phase that is eroded from this last layer
            tmp = (self.stratH[nids, eroLayNb] - eroVal) * self.stratF[nids, eroLayNb]
            tmp[tmp < 1.0e-8] = 0.0
            thFine += tmp * (1.0 - self.phiF[nids, eroLayNb])
            # Define the uncompacted fine thickness that will be deposited dowstream
            self.thFine[nids] = thFine / (1.0 - self.phi0f)
            self.thFine[self.thFine < 0.0] = 0.0

        if self.stratW is not None:
            # Get thickness of each sediment type eroded in the remaining layer
            self.thClay = np.zeros(self.lpoints)
            # From weathered thickness extract the solid phase that is eroded from this last layer
            tmp = (self.stratH[nids, eroLayNb] - eroVal) * self.stratW[nids, eroLayNb]
            tmp[tmp < 1.0e-8] = 0.0
            thClay += tmp * (1.0 - self.phiW[nids, eroLayNb])
            # Define the uncompacted weathered thickness that will be deposited dowstream
            self.thClay[nids] = thClay / (1.0 - self.phi0w)
            self.thClay[self.thClay < 0.0] = 0.0

        self.thCoarse = np.zeros(self.lpoints)
        if self.carbOn:
            # From carb thickness extract the solid phase that is eroded from this last layer
            self.thCarb = np.zeros(self.lpoints)
            tmp = (self.stratH[nids, eroLayNb] - eroVal) * self.stratC[nids, eroLayNb]
            tmp[tmp < 1.0e-8] = 0.0
            thCarb += tmp * (1.0 - self.phiC[nids, eroLayNb])
            # Define the uncompacted carbonate thickness that will be deposited dowstream
            self.thCarb[nids] = thCarb / (1.0 - self.phi0c)
            self.thCarb[self.thCarb < 0.0] = 0.0
            # From sand thickness extract the solid phase that is eroded from this last layer
            tmp = self.stratH[nids, eroLayNb] - eroVal
            tmp *= (
                1.0
                - self.stratC[nids, eroLayNb]
                - self.stratW[nids, eroLayNb]
                - self.stratF[nids, eroLayNb]
            )
            # Define the uncompacted sand thickness that will be deposited dowstream
            thCoarse += tmp * (1.0 - self.phiS[nids, eroLayNb])
            self.thCoarse[nids] = thCoarse / (1.0 - self.phi0s)
            self.thCoarse[self.thCoarse < 0.0] = 0.0
        else:
            # From sand thickness extract the solid phase that is eroded from this last layer
            tmp = self.stratH[nids, eroLayNb] - eroVal
            if self.stratF is not None:
                tmp *= 1.0 - self.stratF[nids, eroLayNb] - self.stratW[nids, eroLayNb]
            tmp[tmp < 1.0e-8] = 0.0
            # Define the uncompacted sand thickness that will be deposited dowstream
            thCoarse += tmp * (1.0 - self.phiS[nids, eroLayNb])
            self.thCoarse[nids] = thCoarse / (1.0 - self.phi0s)
            self.thCoarse[self.thCoarse < 0.0] = 0.0

        # Update thickness of top stratigraphic layer
        self.stratH[nids, eroLayNb] = eroVal
        self.stratH[nids, 0] -= 1.0e6
        self.stratH[self.stratH <= 0] = 0.0
        self.phiS[self.stratH <= 0] = 0.0
        if self.stratF is not None:
            self.stratF[self.stratH <= 0] = 0.0
            self.phiF[self.stratH <= 0] = 0.0
        if self.stratW is not None:
            self.stratW[self.stratH <= 0] = 0.0
            self.phiW[self.stratH <= 0] = 0.0
        if self.carbOn:
            self.stratC[self.stratH <= 0] = 0.0
            self.phiC[self.stratH <= 0] = 0.0

        self.thCoarse /= self.dt
        if self.stratF is not None:
            self.thFine /= self.dt
            del thickF, thFine
        if self.stratW is not None:
            self.thClay /= self.dt
            del thickW, thClay
        if self.carbOn:
            self.thCarb /= self.dt
            del thickC, thCarb

        del ero, nids, cumThick, boolMask, mask, tmp, eroLayNb, eroVal
        del thickS, thCoarse

        gc.collect()

        return

    def elevStrat(self):
        """
        This function updates the current stratigraphic layer elevation.
        """

        self.stratZ[:, self.stratStep] = self.hLocal.getArray()

        return

    def _depthPorosity(self, depth):
        """
        This function uses the depth-porosity relationships to compute the porosities for each lithology and then the solid phase to get each layer thickness changes due to compaction.

        .. note::

            We assume that porosity cannot increase after unloading.

        :arg depth: depth below basement for each sedimentary layer

        :return: newH updated sedimentary layer thicknesses after compaction
        """

        # Depth-porosity functions
        phiS = self.phi0s * np.exp(depth / self.z0s)
        phiS = np.minimum(phiS, self.phiS[:, : self.stratStep + 1])
        if self.stratF is not None:
            phiF = self.phi0f * np.exp(depth / self.z0f)
            phiF = np.minimum(phiF, self.phiF[:, : self.stratStep + 1])
            phiW = self.phi0w * np.exp(depth / self.z0w)
            phiW = np.minimum(phiW, self.phiW[:, : self.stratStep + 1])
        if self.carbOn:
            phiC = self.phi0c * np.exp(depth / self.z0c)
            phiC = np.minimum(phiC, self.phiC[:, : self.stratStep + 1])

        # Compute the solid phase in each layers
        if self.stratF is not None:
            tmpF = (
                self.stratH[:, : self.stratStep + 1]
                * self.stratF[:, : self.stratStep + 1]
            )
            tmpF *= 1.0 - self.phiF[:, : self.stratStep + 1]
            tmpW = (
                self.stratH[:, : self.stratStep + 1]
                * self.stratW[:, : self.stratStep + 1]
            )
            tmpW *= 1.0 - self.phiW[:, : self.stratStep + 1]

        if self.carbOn:
            tmpC = (
                self.stratH[:, : self.stratStep + 1]
                * self.stratC[:, : self.stratStep + 1]
            )
            tmpC *= 1.0 - self.phiC[:, : self.stratStep + 1]
            tmpS = (
                self.stratC[:, : self.stratStep + 1]
                + self.stratF[:, : self.stratStep + 1]
                + self.stratW[:, : self.stratStep + 1]
            )
            tmpS = self.stratH[:, : self.stratStep + 1] * (1.0 - tmpS)
            tmpS *= 1.0 - self.phiS[:, : self.stratStep + 1]
            solidPhase = tmpC + tmpS + tmpF + tmpW
        else:
            if self.stratF is not None:
                tmpS = self.stratH[:, : self.stratStep + 1] * (
                    1.0
                    - self.stratF[:, : self.stratStep + 1]
                    - self.stratW[:, : self.stratStep + 1]
                )
                tmpS *= 1.0 - self.phiS[:, : self.stratStep + 1]
                solidPhase = tmpS + tmpF + tmpW
            else:
                tmpS = self.stratH[:, : self.stratStep + 1]
                tmpS *= 1.0 - self.phiS[:, : self.stratStep + 1]
                solidPhase = tmpS

        # Get new layer thickness after porosity change
        if self.stratF is not None:
            tmpF = self.stratF[:, : self.stratStep + 1] * (
                1.0 - phiF[:, : self.stratStep + 1]
            )
            tmpW = self.stratW[:, : self.stratStep + 1] * (
                1.0 - phiW[:, : self.stratStep + 1]
            )
        if self.carbOn:
            tmpC = self.stratC[:, : self.stratStep + 1] * (
                1.0 - phiC[:, : self.stratStep + 1]
            )
            tmpS = (
                1.0
                - self.stratF[:, : self.stratStep + 1]
                - self.stratC[:, : self.stratStep + 1]
                - self.stratW[:, : self.stratStep + 1]
            )
            tmpS *= 1.0 - phiS[:, : self.stratStep + 1]
            tot = tmpS + tmpC + tmpF + tmpW
        else:
            if self.stratF is not None:
                tmpS = (
                    1.0
                    - self.stratF[:, : self.stratStep + 1]
                    - self.stratW[:, : self.stratStep + 1]
                ) * (1.0 - phiS[:, : self.stratStep + 1])
                tot = tmpS + tmpF + tmpW
            else:
                tot = 1.0 - phiS[:, : self.stratStep + 1]

        ids = np.where(tot > 0.0)
        newH = np.zeros(tot.shape)
        newH[ids] = solidPhase[ids] / tot[ids]
        newH[newH <= 0] = 0.0
        phiS[newH <= 0] = 0.0
        if self.stratF is not None:
            phiF[newH <= 0] = 0.0
            phiW[newH <= 0] = 0.0
        if self.carbOn:
            phiC[newH <= 0] = 0.0

        # Update porosities in each sedimentary layer
        self.phiS[:, : self.stratStep + 1] = phiS
        if self.stratF is not None:
            self.phiF[:, : self.stratStep + 1] = phiF
            self.phiW[:, : self.stratStep + 1] = phiW
        if self.carbOn:
            self.phiC[:, : self.stratStep + 1] = phiC

        del phiS, solidPhase
        del ids, tmpS, tot
        if self.stratF is not None:
            del tmpF, phiF, tmpW, phiW
        if self.carbOn:
            del phiC, tmpC

        gc.collect()

        return newH

    def getCompaction(self):
        """
        This function computes the changes in sedimentary layers porosity and thicknesses due to compaction.

        .. note::

            We assume simple depth-porosiy relationships for each sediment type available in each layers.
        """

        t0 = process_time()
        topZ = np.vstack(self.hLocal.getArray())
        totH = np.sum(self.stratH[:, : self.stratStep + 1], axis=1)

        # Height of the sediment column above the center of each layer is given by
        cumZ = -np.cumsum(self.stratH[:, self.stratStep :: -1], axis=1) + topZ
        elev = np.append(topZ, cumZ[:, :-1], axis=1)
        zlay = np.fliplr(elev - np.fliplr(self.stratH[:, : self.stratStep + 1] / 2.0))

        # Compute lithologies porosities for each depositional layers
        # Get depth below basement
        depth = zlay - topZ

        # Now using depth-porosity relationships we compute the porosities
        newH = self._depthPorosity(depth)

        # Get the total thickness changes induced by compaction and
        # update the elevation accordingly
        dz = totH - np.sum(newH, axis=1)
        dz[dz <= 0] = 0.0
        self.hLocal.setArray(topZ.flatten() - dz.flatten())
        self.dm.localToGlobal(self.hLocal, self.hGlobal)

        # Update each layer thicknesses
        self.stratH[:, : self.stratStep + 1] = newH
        del dz, newH, totH, topZ
        del depth, zlay, cumZ, elev

        gc.collect()

        if MPIrank == 0 and self.verbose:
            print(
                "Compute Lithology Porosity Values (%0.02f seconds)"
                % (process_time() - t0),
                flush=True,
            )

        return

    def updateDispStrata(self):
        """
        Gets the stratigraphic records relevant to each partition after mesh advection.

        The functions returns local stratigraphic layer information:

        - thickness of each stratigrapic layer `loc_stratH` accounting for both erosion & deposition events.
        - proportion of fine sediment `loc_stratF` contains in each stratigraphic layer.
        - proportion of weathered sediment `loc_stratW` contains in each stratigraphic layer.
        - elevation at time of deposition, considered to be to the current elevation for the top stratigraphic layer `loc_stratZ`.
        - porosity of coarse sediment `loc_phiS` in each stratigraphic layer computed at center of each layer.
        - porosity of fine sediment `loc_phiF` in each stratigraphic layer computed at center of each layer.
        - porosity of waethered sediment `loc_phiW` in each stratigraphic layer computed at center of each layer.
        - proportion of carbonate sediment `loc_strataC` contains in each stratigraphic layer if the carbonate module is turned on.
        - porosity of carbonate sediment `loc_phiC` in each stratigraphic layer computed at center of each layer when the carbonate module is turned on.

        .. note::

            In `gospl`, the stratigraphic layers are only defined locally. For interpolation on the edges of each partition it is important to ensure that all stratigraphic information on the adjacent nodes of the neighbouring partition are accessible. This is done by applying MPI `Allreduce` operation to the nodes parts of the overlaid ('shadow') zone.

        :return: loc_stratH, loc_stratZ, loc_stratF, loc_stratW, loc_stratC, loc_phiS, loc_phiF, loc_phiW, loc_phiC
        """

        stratH = self.stratH[self.lgIDs, : self.stratStep]
        if MPIsize > 1:
            temp = np.full((self.shadowgNb, self.stratStep), -1.0e8, dtype=np.float64)
            temp[self.gshadinIDs, :] = stratH[self.gshadowIDs, :]
            temp[self.gshadoutIDs, :] = -1.0e8
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
            stratH[self.shadowAlls, :] = temp
            loc_stratH = stratH[self.locIDs, :]
        else:
            loc_stratH = stratH[self.locIDs, :]

        stratZ = self.stratZ[self.lgIDs, : self.stratStep]
        if MPIsize > 1:
            temp.fill(-1.0e8)
            temp[self.gshadinIDs, :] = stratZ[self.gshadowIDs, :]
            temp[self.gshadoutIDs, :] = -1.0e8
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
            stratZ[self.shadowAlls, :] = temp
            loc_stratZ = stratZ[self.locIDs, :]
        else:
            loc_stratZ = stratZ[self.locIDs, :]

        phiS = self.phiS[self.lgIDs, : self.stratStep]
        if MPIsize > 1:
            temp.fill(-1.0e8)
            temp[self.gshadinIDs, :] = phiS[self.gshadowIDs, :]
            temp[self.gshadoutIDs, :] = -1.0e8
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
            phiS[self.shadowAlls, :] = temp
            loc_phiS = phiS[self.locIDs, :]
        else:
            loc_phiS = phiS[self.locIDs, :]

        if self.stratF is not None:
            stratF = self.stratF[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = stratF[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                stratF[self.shadowAlls, :] = temp
                loc_stratF = stratF[self.locIDs, :]
            else:
                loc_stratF = stratF[self.locIDs, :]

            phiF = self.phiF[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = phiF[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                phiF[self.shadowAlls, :] = temp
                loc_phiF = phiF[self.locIDs, :]
            else:
                loc_phiF = phiF[self.locIDs, :]
        else:
            loc_stratF = None
            loc_phiF = None

        if self.stratW is not None:
            stratW = self.stratW[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = stratW[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                stratW[self.shadowAlls, :] = temp
                loc_stratW = stratW[self.locIDs, :]
            else:
                loc_stratW = stratW[self.locIDs, :]

            phiW = self.phiW[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = phiW[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                phiW[self.shadowAlls, :] = temp
                loc_phiW = phiW[self.locIDs, :]
            else:
                loc_phiW = phiW[self.locIDs, :]
        else:
            loc_stratW = None
            loc_phiW = None

        if self.carbOn:
            stratC = self.stratC[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = stratZ[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                stratC[self.shadowAlls, :] = temp
                loc_stratC = stratC[self.locIDs, :]
            else:
                loc_stratC = stratC[self.locIDs, :]

            phiC = self.phiC[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = phiC[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                phiC[self.shadowAlls, :] = temp
                loc_phiC = phiC[self.locIDs, :]
            else:
                loc_phiC = phiC[self.locIDs, :]

            return (
                loc_stratH,
                loc_stratZ,
                loc_stratF,
                loc_stratW,
                loc_stratC,
                loc_phiS,
                loc_phiF,
                loc_phiW,
                loc_phiC,
            )
        else:
            return (
                loc_stratH,
                loc_stratZ,
                loc_stratF,
                loc_stratW,
                loc_phiS,
                loc_phiF,
                loc_phiW,
            )

    def stratalRecord(
        self,
        indices,
        weights,
        loc_stratH,
        loc_stratZ,
        loc_stratF,
        loc_stratW,
        loc_stratC,
        loc_phiS,
        loc_phiF,
        loc_phiW,
        loc_phiC,
    ):
        """
        Once the interpolation has been performed, the following function updates the stratigraphic records based on the advected mesh.

        The function relies on 3 fortran subroutines (for loop performance purposes):

        1. stratasimple
        2. stratabuild
        3. stratabuildcarb

        :arg indices: indices of the closest nodes used for interpolation
        :arg weights: weights based on the distances to closest nodes
        :arg loc_stratH: thickness of each stratigrapic layer accounting for both erosion & deposition events.
        :arg loc_stratF: proportion of fine sediment contains in each stratigraphic layer.
        :arg loc_stratW: proportion of weathered sediment contains in each stratigraphic layer.
        :arg loc_stratZ: elevation at time of deposition, considered to be to the current elevation for the top stratigraphic layer.
        :arg loc_phiS: porosity of coarse sediment in each stratigraphic layer computed at center of each layer.
        :arg loc_phiF: porosity of fine sediment in each stratigraphic layer computed at center of each layer.
        :arg loc_phiW: porosity of weathered sediment in each stratigraphic layer computed at center of each layer.
        :arg loc_strataC: proportion of carbonate sediment contains in each stratigraphic layer if the carbonate module is turned on.
        :arg loc_phiC: porosity of carbonate sediment in each stratigraphic layer computed at center of each layer when the carbonate module is turned on.

        """

        if self.carbOn:
            (
                self.stratH[:, : self.stratStep],
                self.stratZ[:, : self.stratStep],
                self.stratF[:, : self.stratStep],
                self.stratW[:, : self.stratStep],
                self.stratC[:, : self.stratStep],
                self.phiS[:, : self.stratStep],
                self.phiF[:, : self.stratStep],
                self.phiW[:, : self.stratStep],
                self.phiC[:, : self.stratStep],
            ) = stratabuildcarb(
                self.lpoints,
                self.stratStep,
                indices,
                weights,
                loc_stratH,
                loc_stratZ,
                loc_stratF,
                loc_stratW,
                loc_stratC,
                loc_phiS,
                loc_phiF,
                loc_phiW,
                loc_phiC,
            )
        elif self.stratF is not None:
            (
                self.stratH[:, : self.stratStep],
                self.stratZ[:, : self.stratStep],
                self.stratF[:, : self.stratStep],
                self.stratW[:, : self.stratStep],
                self.phiS[:, : self.stratStep],
                self.phiF[:, : self.stratStep],
                self.phiW[:, : self.stratStep],
            ) = stratabuild(
                self.lpoints,
                self.stratStep,
                indices,
                weights,
                loc_stratH,
                loc_stratZ,
                loc_stratF,
                loc_stratW,
                loc_phiS,
                loc_phiF,
                loc_phiW,
            )
        else:
            (
                self.stratH[:, : self.stratStep],
                self.stratZ[:, : self.stratStep],
                self.phiS[:, : self.stratStep],
            ) = stratasimple(
                self.lpoints,
                self.stratStep,
                indices,
                weights,
                loc_stratH,
                loc_stratZ,
                loc_phiS,
            )

        return

    def localStrat(self):
        """
        Updates stratigraphic records after mesh advection on the edges of each partition to ensure that all stratigraphic information on the adjacent nodes of the neighbouring partition are equals on all processors sharing a common number of nodes.
        """

        stratH = self.stratH[self.lgIDs, : self.stratStep]
        if MPIsize > 1:
            temp = np.full((self.shadowgNb, self.stratStep), -1.0e8, dtype=np.float64)
            temp[self.gshadinIDs, :] = stratH[self.gshadowIDs, :]
            temp[self.gshadoutIDs, :] = -1.0e8
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
            stratH[self.shadowAlls, :] = temp
            self.stratH[:, : self.stratStep] = stratH[self.locIDs, :]
        else:
            self.stratH[:, : self.stratStep] = stratH[self.locIDs, :]
        stratZ = self.stratZ[self.lgIDs, : self.stratStep]
        if MPIsize > 1:
            temp.fill(-1.0e8)
            temp[self.gshadinIDs, :] = stratZ[self.gshadowIDs, :]
            temp[self.gshadoutIDs, :] = -1.0e8
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
            stratZ[self.shadowAlls, :] = temp
            self.stratZ[:, : self.stratStep] = stratZ[self.locIDs, :]
        else:
            self.stratZ[:, : self.stratStep] = stratZ[self.locIDs, :]
        phiS = self.phiS[self.lgIDs, : self.stratStep]
        if MPIsize > 1:
            temp.fill(-1.0e8)
            temp[self.gshadinIDs, :] = phiS[self.gshadowIDs, :]
            temp[self.gshadoutIDs, :] = -1.0e8
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
            phiS[self.shadowAlls, :] = temp
            self.phiS[:, : self.stratStep] = phiS[self.locIDs, :]
        else:
            self.phiS[:, : self.stratStep] = phiS[self.locIDs, :]
        del stratH, stratZ, phiS

        if self.stratF is not None:
            stratF = self.stratF[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = stratF[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                stratF[self.shadowAlls, :] = temp
                self.stratF[:, : self.stratStep] = stratF[self.locIDs, :]
            else:
                self.stratF[:, : self.stratStep] = stratF[self.locIDs, :]
            phiF = self.phiF[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = phiF[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                phiF[self.shadowAlls, :] = temp
                self.phiF[:, : self.stratStep] = phiF[self.locIDs, :]
            else:
                self.phiF[:, : self.stratStep] = phiF[self.locIDs, :]
            del stratF, phiF

        if self.stratW is not None:
            stratW = self.stratW[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = stratW[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                stratW[self.shadowAlls, :] = temp
                self.stratW[:, : self.stratStep] = stratW[self.locIDs, :]
            else:
                self.stratW[:, : self.stratStep] = stratW[self.locIDs, :]
            phiW = self.phiW[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = phiW[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                phiW[self.shadowAlls, :] = temp
                self.phiW[:, : self.stratStep] = phiW[self.locIDs, :]
            else:
                self.phiW[:, : self.stratStep] = phiW[self.locIDs, :]
            del stratW, phiW

        if self.carbOn:
            stratC = self.stratC[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = stratC[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                stratC[self.shadowAlls, :] = temp
                self.stratC[:, : self.stratStep] = stratC[self.locIDs, :]
            else:
                self.stratC[:, : self.stratStep] = stratC[self.locIDs, :]
            phiC = self.phiC[self.lgIDs, : self.stratStep]
            if MPIsize > 1:
                temp.fill(-1.0e8)
                temp[self.gshadinIDs, :] = phiC[self.gshadowIDs, :]
                temp[self.gshadoutIDs, :] = -1.0e8
                MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, temp, op=MPI.MAX)
                phiC[self.shadowAlls, :] = temp
                self.phiC[:, : self.stratStep] = phiC[self.locIDs, :]
            else:
                self.phiC[:, : self.stratStep] = phiC[self.locIDs, :]
            del stratC, phiC

        return



import os
import gc
import sys
import vtk
import warnings
import petsc4py
import numpy as np
from scipy import spatial
import numpy_indexed as npi

from mpi4py import MPI
from time import process_time
from vtk.util import numpy_support

if "READTHEDOCS" not in os.environ:
    from gospl._fortran import fillpit
    from gospl._fortran import mfdreceivers

petsc4py.init(sys.argv)
MPIrank = petsc4py.PETSc.COMM_WORLD.Get_rank()
MPIcomm = petsc4py.PETSc.COMM_WORLD


class FAMesh(object):
    """
    This class calculates **drainage area** in an implicit, iterative manner using PETSc solvers. It accounts  for multiple flow direction paths (SFD to MFD) based on user input declaration.

    .. note::

        The class follows the parallel approach described in `Richardson et al., 2014 <https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2013WR014326>`_ where the iterative nature of the computational algorithms used to solve the linear system creates the possibility of accelerating the solution by providing an initial guess.

    For drainage computation, the class requires to compute depressionless surfaces and the *priority-flood + ϵ* variant of the algorithm proposed in `Barnes et al. (2014) <https://doi.org/10.1016/j.cageo.2013.04.024>`_ is used. It provides a solution to remove flat surfaces, and it produces surfaces for which each cell has a defined gradient from which flow directions can be determined.

    Finally, the class computes river incision expressed using a **stream power formulation** function of river discharge and slope.

    """

    def __init__(self, *args, **kwargs):
        """
        The initialisation of `FAMesh` class consists in the declaration of PETSc vectors and matrices.
        """

        # KSP solver parameters
        self.rtol = 1.0e-8

        # Identity matrix construction
        self.II = np.arange(0, self.lpoints + 1, dtype=petsc4py.PETSc.IntType)
        self.JJ = np.arange(0, self.lpoints, dtype=petsc4py.PETSc.IntType)
        self.iMat = self._matrix_build_diag(np.ones(self.lpoints))

        # Empty matrix construction
        self.II = np.arange(0, self.lpoints + 1, dtype=petsc4py.PETSc.IntType)
        self.JJ = np.arange(0, self.lpoints, dtype=petsc4py.PETSc.IntType)
        self.zMat = self._matrix_build_diag(np.zeros(self.lpoints))

        # Petsc vectors
        # self.fillFAG = self.hGlobal.duplicate()
        self.fillFAL = self.hLocal.duplicate()
        self.FAG = self.hGlobal.duplicate()
        self.FAL = self.hLocal.duplicate()
        self.hOld = self.hGlobal.duplicate()
        self.hOldLocal = self.hLocal.duplicate()
        self.Eb = self.hGlobal.duplicate()
        self.stepED = self.hGlobal.duplicate()
        self.EbLocal = self.hLocal.duplicate()

        # Clinoforms slopes for each sediment type
        self.slps = [6.0, 5.0, 3.0, 1.0]

        return

    def _matrix_build(self, nnz=(1, 1)):
        """
        Creates a sparse PETSc matrix.

        .. note::

            To achieve good performance during matrix assembly, the function preallocates the matrix storage by setting the array nnz.

        :arg nnz: array containing the number of nonzeros in the various rows

        :return: sparse PETSc matrix
        """

        matrix = petsc4py.PETSc.Mat().create(comm=MPIcomm)
        matrix.setType("aij")
        matrix.setSizes(self.sizes)
        matrix.setLGMap(self.lgmap_row, self.lgmap_col)
        matrix.setFromOptions()
        matrix.setPreallocationNNZ(nnz)

        return matrix

    def _matrix_build_diag(self, V, nnz=(1, 1)):
        """
        Builds a PETSc diagonal matrix based on a given array `V`

        :arg V: diagonal data array
        :arg nnz: array containing the number of nonzero blocks

        :return: sparse PETSc matrix
        """

        matrix = self._matrix_build()

        # Define diagonal matrix
        matrix.assemblyBegin()
        matrix.setValuesLocalCSR(
            self.II,
            self.JJ,
            V,
        )
        matrix.assemblyEnd()

        return matrix

    def _make_reasons(self, reasons):
        """
        Provides reasons for PETSc error...
        """

        return dict(
            [(getattr(reasons, r), r) for r in dir(reasons) if not r.startswith("_")]
        )

    def _solve_KSP(self, guess, matrix, vector1, vector2):
        """

        PETSc *scalable linear equations solvers* (**KSP**) component provides Krylov subspace iterative method and a preconditioner. Here, flow accumulation solution is obtained using PETSc Richardson solver (`richardson`) with block Jacobian preconditioning (`bjacobi`).

        .. note::

            The solver choice was made based on the convergence results from `Richardson et al. (2014) <https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2013WR014326>`_ but can be changed if better solver and preconditioner combinations are found.

        Using such iterative method allows for an initial guess to be provided. When this initial guess is close to the solution, the number of iterations required for convergence dramatically decreases. Here the flow discharge solution from previous time step can be passed as an initial `guess` to the solver as discharge often exhibits little change between successive time intervals.

        :arg guess: Boolean specifying if the iterative KSP solver initial guess is nonzero (when provided it corresponds to the previous flow discharge values)
        :arg matrix: PETSc sparse matrix used by the KSP solver composed of diagonal terms set to unity (identity matrix) and off-diagonal terms (weights between 0 and 1). The weights are calculated based on the number of downslope neighbours (based on the chosen number of flow direction directions) and are proportional to the slope.
        :arg vector1: PETSc vector corresponding to the local volume of water available for runoff during a given time step (*e.g.* voronoi area times local precipitation rate)
        :arg vector2: PETSc vector corresponding to the unknown flow discharge values

        :return: vector2 PETSc vector of the new flow discharge values
        """

        ksp = petsc4py.PETSc.KSP().create(petsc4py.PETSc.COMM_WORLD)
        if guess:
            ksp.setInitialGuessNonzero(guess)
        ksp.setOperators(matrix, matrix)
        ksp.setType("richardson")
        pc = ksp.getPC()
        pc.setType("bjacobi")
        ksp.setTolerances(rtol=self.rtol)
        ksp.solve(vector1, vector2)
        r = ksp.getConvergedReason()
        if r < 0:
            KSPReasons = self._make_reasons(petsc4py.PETSc.KSP.ConvergedReason())
            print(
                "LinearSolver failed to converge after %d iterations",
                ksp.getIterationNumber(),
                flush=True,
            )
            print("with reason: %s", KSPReasons[r], flush=True)
            raise RuntimeError("LinearSolver failed to converge!")
        ksp.destroy()

        return vector2

    def _buildFlowDirection(self, h):
        """
        This function builds from neighbouring slopes the flow directions. It calls a fortran subroutine that locally computes for each vertice:

        - the indices of receivers (downstream) nodes depending on the desired number of flow directions (SFD to MFD).
        - the distances to the receivers based on mesh resolution.
        - the associated weights calculated based on the number of receivers and proportional to the slope.

        :arg h: elevation numpy array
        """

        t0 = process_time()

        # Define multiple flow directions for unfilled elevation
        self.rcvID, self.distRcv, self.wghtVal = mfdreceivers(
            self.flowDir, self.flowExp, self.inIDs, h, self.sealevel
        )
        # Get marine regions
        self.seaID = np.where(h <= self.sealevel)[0]

        # Set marine nodes
        if self.flatModel:
            self.rcvID[self.idBorders, :] = np.tile(self.idBorders, (self.flowDir, 1)).T
            self.distRcv[self.idBorders, :] = 0.0
            self.wghtVal[self.idBorders, :] = 0.0

        # Get global nodes that have no receivers
        sum_weight = np.sum(self.wghtVal, axis=1)
        tmpw = np.zeros(self.mpoints, dtype=np.float64)
        tmpw[self.locIDs] = sum_weight
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, tmpw, op=MPI.MAX)
        self.pitPts = tmpw == 0.0

        if MPIrank == 0 and self.verbose:
            print(
                "Flow Direction declaration (%0.02f seconds)" % (process_time() - t0),
                flush=True,
            )

        return

    def _distanceCoasts(self, data, k_neighbors=1):
        """
        This function computes for every marine vertices the distance to the closest coastline.

        .. important::

            The calculation takes advantage of the `vtkContourFilter` function from VTK library
            which is performed on the **global** VTK mesh. Once the coastlines have been extracted,
            the distances are obtained by querying a kd-tree (initialised with the coastal nodes) for
            marine vertices contained within each partition.

        :arg data: global elevation numpy array
        :arg k_neighbors: number of nodes to use when querying the kd-tree
        """

        t0 = process_time()

        self.coastDist = np.zeros(self.lpoints)
        pointData = self.vtkMesh.GetPointData()
        array = numpy_support.numpy_to_vtk(data, deep=1)
        array.SetName("elev")
        pointData.AddArray(array)
        self.vtkMesh.SetFieldData(pointData)

        cf = vtk.vtkContourFilter()
        cf.SetInputData(self.vtkMesh)
        cf.SetValue(0, self.sealevel)
        cf.SetInputArrayToProcess(0, 0, 0, 0, "elev")
        cf.GenerateTrianglesOff()
        cf.Update()
        coastXYZ = numpy_support.vtk_to_numpy(cf.GetOutput().GetPoints().GetData())
        tree = spatial.cKDTree(coastXYZ, leafsize=10)
        self.coastDist[self.seaID], indices = tree.query(
            self.lcoords[self.seaID, :], k=k_neighbors
        )
        del array, pointData, cf, tree, indices, coastXYZ
        gc.collect()

        if MPIrank == 0 and self.verbose:
            print(
                "Construct Distance to Coast (%0.02f seconds)" % (process_time() - t0),
                flush=True,
            )

        return

    def _flowSurface(self):
        """
        This function computes the depression less surface.

        .. note::

            In most landscape evolution models, internally draining regions (*e.g.* depressions and pits) are usually filled before the calculation of flow discharge and erosion–deposition rates. This ensures that all flows conveniently reach the coast or the boundary of the simulated domain. In models intended to simulate purely erosional features, such depressions are usually treated as transient features and often ignored.

            However, `gospl` is designed to not only address erosion problems but also to simulate source-to-sink transfer and sedimentary basins formation and evolution in potentially complex tectonic settings. In such cases, depressions may be formed at different periods during runtime and may be filled or remain internally drained (*e.g.* endorheic basins) depending on the volume of sediment transported by upstream catchments.

        The function is **not parallelised** and is performed on the master processor. It calls a fortran subroutine `fillpit` that uses a *priority-flood + ϵ* variant of the algorithm proposed in `Barnes et al. (2014) <https://doi.org/10.1016/j.cageo.2013.04.024>`_ for unstructured meshes.

        The initialisation step consists of pushing all the marine nodes which are neighbours to a deep ocean node (*i.e.* nodes at water depth below 8000 m) onto a priority queue. The priority queue rearranges these nodes so that the ones with the lowest elevations in the queue are always processed first. The algorithm then proceeds incrementally by adding new vertices in the queue and defines filled elevations in the ascending order of their spill elevations + ϵ.

        At the end of the function, the global fill-limited elevations are returned as well as the depressions charcateristics (*i.e.* a unique indice for each depressions, their volumes and the positions of the spill-over global vertices).

        :return: gZ hFill (global and filled elevations numpy arrays)
        """

        # Get global elevations for pit filling...
        t0 = process_time()
        hl = self.hLocal.getArray().copy()
        gZ = np.zeros(self.mpoints, dtype=np.float64)
        gZ[self.locIDs] = hl
        gZ[self.outIDs] = -1.0e8
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, gZ, op=MPI.MAX)

        # Perform pit filling on process rank 0
        if MPIrank == 0:
            # fillpit returns:
            # - hFill: filled elevation values
            # - pits: 2D array containing each pit ID and
            #         corresponding overspilling point ID
            hFill, _ = fillpit(self.sealevel - 1.0, gZ, self.waterfill)

        else:
            hFill = np.zeros(self.mpoints, dtype=np.float64)
        hFill = MPI.COMM_WORLD.bcast(hFill, root=0)
        self.hFill = hFill[self.locIDs]

        if MPIrank == 0 and self.verbose:
            print(
                "Get pit filling information (%0.02f seconds)" % (process_time() - t0),
                flush=True,
            )

        del hl
        gc.collect()

        return gZ, hFill

    def matrixFlow(self, flow=True):
        """
        This function defines the flow direction matrices for filled-limited elevations.

        .. note::

            The matrix is built incrementally looping through the number of flow direction paths defined by the user. It proceeds by assembling a local Compressed Sparse Row (**CSR**) matrix to a global PETSc matrix.

            When setting up the flow matrix in PETSc, we preallocate the non-zero entries of the matrix before starting filling in the values. Using PETSc sparse matrix storage scheme has the advantage that matrix-vector multiplication is extremely fast.

        The  matrix coefficients consist of weights (comprised between 0 and 1) calculated based on the number of downslope neighbours and proportional to the slope.

        :arg flow: boolean to compute matrix for either downstream water or sediment transport

        """

        flowMat = self.iMat.copy()
        indptr = np.arange(0, self.lpoints + 1, dtype=petsc4py.PETSc.IntType)
        nodes = indptr[:-1]

        if flow:
            wght = self.wghtVal
            rcv = self.rcvID
        else:
            wght = self.lWght
            rcv = self.lRcv

        for k in range(0, self.flowDir):

            # Flow direction matrix for a specific direction
            tmpMat = self._matrix_build()
            data = -wght[:, k].copy()
            data[rcv[:, k].astype(petsc4py.PETSc.IntType) == nodes] = 0.0
            tmpMat.assemblyBegin()
            tmpMat.setValuesLocalCSR(
                indptr,
                rcv[:, k].astype(petsc4py.PETSc.IntType),
                data,
            )
            tmpMat.assemblyEnd()

            # Add the weights from each direction
            flowMat += tmpMat
            tmpMat.destroy()

        del data, indptr, nodes
        gc.collect()

        # Store flow accumulation matrix
        self.fMat = flowMat.transpose().copy()

        flowMat.destroy()

        return

    def _distributeFlowExcess(self):
        """
        In cases where rivers flow in depressions, they might fill the sink completely and overspill or a portion of it forming a lake. This function computes the lake elevation and the excess of water (if any) able to flow dowstream.

        .. important::

            The excess water is then added to the downstream flow accumulation (`FA`) and used to estimate rivers' erosion.

        """

        t0 = process_time()
        self.hierarchicalSink(True, True, False, None)
        lFA = self.FAL.getArray().copy()
        nFA = np.zeros(self.mpoints, dtype=np.float64)
        nFA[self.locIDs] = lFA
        nFA[self.outIDs] = 0.0
        MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, nFA, op=MPI.MAX)
        FA = np.zeros(self.mpoints, dtype=np.float64)
        FA[self.pitPts] = nFA[self.pitPts]
        FA[self.sinkID] = 0.0
        if self.flatModel:
            FA[self.glBorders] = 0.0
        self.depo = np.zeros(self.mpoints, dtype=np.float64)
        inFA = np.where(FA > 0)[0]

        if len(inFA) == 0:
            del lFA, FA, nFA, inFA
            gc.collect()
            return

        step = 0
        self.flowDist = True
        while (FA > 0).any():

            # Check if FA are in closed sea
            insea = np.zeros(1, dtype=np.int64)
            if MPIrank == 0:
                if (self.sIns[inFA] > -1).any():
                    insea[0] = 1
            MPI.COMM_WORLD.Allreduce(MPI.IN_PLACE, insea, op=MPI.MAX)

            if insea[0] > 0:
                # Distribute sediments in closed sea
                FA = self.distributeSea(FA)
                # Remove flux that flow into open sea
                FA[self.sinkID] = 0.0
                # In case of 2D model remove flow on mesh borders
                if self.flatModel:
                    FA[self.glBorders] = 0.0

            # At this point excess water are all in continental regions
            if (FA > 0).any():
                FA = self.distributeContinent(FA, None)
                if self.flatModel:
                    FA[self.glBorders] = 0.0
            step += 1
            if step > 20:
                break

        # Store lake water volume for each node
        self.fillFAL.setArray(self.depo[self.locIDs] * self.marea[self.locIDs])
        self.flowDist = False
        if MPIrank == 0 and self.verbose:
            print(
                "Distribute Excess Flow Downstream (%0.02f seconds)"
                % (process_time() - t0),
                flush=True,
            )

        del FA, nFA, inFA, lFA
        gc.collect()

        return

    def flowAccumulation(self):
        """
        This function is the **main entry point** for flow accumulation computation.

        .. note::

            Flow accumulation (`FA`) calculations are a core component of landscape evolution models as they are often used as proxy to estimate flow discharge, sediment load, river width, bedrock erosion, and sediment deposition. Until recently, conventional `FA` algorithms were serial and limited to small spatial problems.

        `gospl` model computes the flow discharge from `FA` and the net precipitation rate using a **parallel implicit drainage area (IDA) method** proposed by `Richardson et al., 2014 <https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2013WR014326>`_ but adapted to unstructured grids.

        It calls the following *private functions*:

        1. _flowSurface
        2. _buildFlowDirection
        3. _distanceCoasts
        4. _solve_KSP
        5. _distributeFlowExcess

        """

        # self.dm.localToGlobal(self.hLocal, self.hGlobal)
        self.dm.globalToLocal(self.hGlobal, self.hLocal)

        # Filled-limited surface used to move river flow downstream
        gZ, hFill = self._flowSurface()

        # Build flow direction
        self._buildFlowDirection(hFill[self.locIDs])

        # Define coastal distance for marine points
        if self.vtkMesh is not None:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                self._distanceCoasts(gZ)

        del hFill, gZ
        gc.collect()

        # Build transport direction matrices
        t0 = process_time()
        self.matrixFlow(True)

        # Solve flow accumulation for filled-limited elevation
        if self.tNow == self.rStart:
            self._solve_KSP(False, self.fMat, self.bG, self.FAG)
        else:
            self._solve_KSP(True, self.fMat, self.bG, self.FAG)
        self.dm.globalToLocal(self.FAG, self.FAL)

        if MPIrank == 0 and self.verbose:
            print(
                "Compute Flow Accumulation (%0.02f seconds)" % (process_time() - t0),
                flush=True,
            )

        # Distribute excess flow
        self._distributeFlowExcess()

        return

    def _getErosionRate(self):
        r"""
        This function computes erosion rates in metres per year. This is done on the filled elevation. We use the filled-limited elevation to ensure that erosion is not going to be underestimate by small depressions which are likely to be filled (either by sediments or water) during a single time step.

        The simplest law to simulate fluvial incision is based on the detachment-limited stream power law, in which erosion rate  depends on drainage area :math:`A`, net precipitation :math:`P` and local slope :math:`S` and takes the form:

        .. math::

          E = − \kappa P^d (PA)^m S^n

        :math:`\kappa` is a dimensional coefficient describing the erodibility of the channel bed as a function of rock strength, bed roughness and climate, :math:`d`, :math:`m` and :math:`n` are dimensionless positive constants.

        Default formulation assumes :math:`d = 0`, :math:`m = 0.5` and :math:`n = 1`. The precipitation exponent :math:`d` allows for representation of climate-dependent chemical weathering of river bed across non-uniform rainfall.

        .. important::

            In `gospl`, the coefficients `m` and `n` are fixed and the only variables that the user can tune are the coefficient `d` and the erodibility :math:`\kappa`. E is in m/yr and therefore the erodibility dimension is :math:`(m yr)^{−0.5}`.

        The erosion rate is solved by an implicit time integration method, the matrix system is based on the receiver distributions and is assembled from local Compressed Sparse Row (**CSR**) matrices into a global PETSc matrix. The PETSc *scalable linear equations solvers* (**KSP**) is used with both an iterative method and a preconditioner and erosion rate solution is obtained using PETSc Richardson solver (`richardson`) with block Jacobian preconditioning (`bjacobi`).

        """

        # Upstream-averaged mean annual precipitation rate and the drainage area
        PA = self.FAL.getArray()

        # Incorporate the effect of local mean annual precipitation rate on erodibility
        Kbr = self.K * (self.rainVal ** self.coeffd)
        Kbr *= np.sqrt(PA) * self.dt
        Kbr[self.seaID] = 0.0

        # Initialise identity matrices...
        EbedMat = self.iMat.copy()
        wght = self.wghtVal.copy()

        # Define erosion coefficients
        for k in range(0, self.flowDir):

            indptr = np.arange(0, self.lpoints + 1, dtype=petsc4py.PETSc.IntType)
            nodes = indptr[:-1]
            # Define erosion limiter to prevent formation of flat
            dh = self.hOldArray - self.hOldArray[self.rcvID[:, k]]
            # Set places which have been filled (e.g. depression/lakes) as no slope regions
            # dh[self.hFill > self.hOldArray] = 0.0
            # dh[dh < 0.0] = 0.0
            limiter = np.divide(dh, dh + 1.0e-3, out=np.zeros_like(dh), where=dh != 0)

            # Bedrock erosion processes SPL computation (maximum bedrock incision)
            data = np.divide(
                Kbr * limiter,
                self.distRcv[:, k],
                out=np.zeros_like(PA),
                where=self.distRcv[:, k] != 0,
            )
            tmpMat = self._matrix_build()
            wght[self.seaID, k] = 0.0
            data = np.multiply(data, -wght[:, k])
            data[self.rcvID[:, k].astype(petsc4py.PETSc.IntType) == nodes] = 0.0
            tmpMat.assemblyBegin()
            tmpMat.setValuesLocalCSR(
                indptr,
                self.rcvID[:, k].astype(petsc4py.PETSc.IntType),
                data,
            )
            tmpMat.assemblyEnd()
            EbedMat += tmpMat
            EbedMat -= self._matrix_build_diag(data)
            tmpMat.destroy()

        del dh, limiter, wght, data
        gc.collect()

        # Solve bedrock erosion thickness
        self._solve_KSP(True, EbedMat, self.hOld, self.stepED)
        EbedMat.destroy()
        self.tmp.waxpy(-1.0, self.hOld, self.stepED)

        # Define erosion rate (positive for incision)
        E = -self.tmp.getArray().copy()
        E = np.divide(E, self.dt)
        E[E < 0.0] = 0.0
        self.Eb.setArray(E)
        self.dm.globalToLocal(self.Eb, self.EbLocal)
        E = self.EbLocal.getArray().copy()

        E[self.seaID] = 0.0
        ids = np.where(
            (self.hOldArray > self.sealevel + 1.0e-2)
            & (self.hOldArray - E * self.dt < self.sealevel + 1.0e-2)
        )[0]
        E[ids] = (self.hOldArray[ids] - self.sealevel - 1.0e-2) / self.dt
        if self.flatModel:
            E[self.idBorders] = 0.0
        # E[self.hFill > self.hOldArray] = 0.0
        self.EbLocal.setArray(E)
        self.dm.localToGlobal(self.EbLocal, self.Eb)

        del E, PA, Kbr, ids
        gc.collect()

        return

    def riverIncision(self):
        """
        River incision is based on a standard form of the **stream power law** assuming
        **detachment-limited behaviour**.

        It calls the private function `_getErosionRate <https://gospl.readthedocs.io/en/latest/api.html#flow.flowplex.FAMesh._getErosionRate>`_ described above. Once erosion rates have been calculated, the function computes local eroded thicknesses for the considered time step and update local elevation and cumulative erosion, deposition values.

        If multiple lithologies are considered, the stratigraphic record is updated based on eroded thicknesses.

        .. important::

            The approach assumes that the volume of rock eroded using the stream power law accounts for both the solid and void phase.

        """

        t0 = process_time()

        # Local & global vectors/arrays
        self.Eb.set(0.0)
        self.hGlobal.copy(result=self.hOld)
        self.dm.globalToLocal(self.hOld, self.hOldLocal)
        self.hOldArray = self.hOldLocal.getArray().copy()

        self._getErosionRate()

        # Get eroded thicknesses
        Eb = self.Eb.getArray().copy()
        self.tmp.setArray(-Eb * self.dt)
        self.cumED.axpy(1.0, self.tmp)
        self.dm.globalToLocal(self.cumED, self.cumEDLocal)

        self.hGlobal.axpy(1.0, self.tmp)
        self.dm.globalToLocal(self.hGlobal, self.hLocal)

        # Update stratigraphic layers
        if self.stratNb > 0:
            self.erodeStrat()

        if MPIrank == 0 and self.verbose:
            print(
                "Get Erosion Thicknesses (%0.02f seconds)" % (process_time() - t0),
                flush=True,
            )

        del Eb
        gc.collect()

        return
